"""Owner related models for Structures."""

# pylint: disable = duplicate-code,too-many-lines

import datetime as dt
import json
import os
import re
from email.utils import format_datetime, parsedate_to_datetime
from typing import Any, Iterable, List, Optional

from bravado.exception import HTTPForbidden, HTTPNotFound

from django.contrib.auth.models import Group, User
from django.core.exceptions import ObjectDoesNotExist
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models, transaction
from django.db.models import F
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from esi.errors import TokenError
from esi.models import Token
from eveuniverse.models import EveMoon, EvePlanet, EveSolarSystem, EveType

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.allianceauth import notify_admins
from app_utils.datetime import DATETIME_FORMAT
from app_utils.helpers import chunks
from app_utils.logging import LoggerAddTag

from structures import __title__
from structures.app_settings import (
    STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED,
    STRUCTURES_DEVELOPER_MODE,
    STRUCTURES_ESI_DIRECTOR_ERROR_MAX_RETRIES,
    STRUCTURES_FEATURE_CUSTOMS_OFFICES,
    STRUCTURES_FEATURE_SKYHOOKS,
    STRUCTURES_FEATURE_STARBASES,
    STRUCTURES_HOURS_UNTIL_STALE_NOTIFICATION,
    STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES,
    STRUCTURES_NOTIFICATIONS_ARCHIVING_ENABLED,
    STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES,
)
from structures.constants import EveGroupId, EveTypeId
from structures.managers import OwnerManager
from structures.providers import esi

from .eveuniverse import EveSovereigntyMap
from .notifications import (
    EveEntity,
    GeneratedNotification,
    Notification,
    NotificationType,
    Webhook,
)
from .structures_1 import Structure, StructureItem
from .structures_2 import PocoDetails, StarbaseDetail, StarbaseDetailFuel

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class General(models.Model):
    """Meta model for global app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app and view public pages"),
            ("view_corporation_structures", "Can view corporation structures"),
            ("view_alliance_structures", "Can view alliance structures"),
            ("view_all_structures", "Can view all structures"),
            ("add_structure_owner", "Can add new structure owner"),
            (
                "view_all_unanchoring_status",
                "Can view unanchoring timers for all structures the user can see",
            ),
            ("view_structure_fit", "Can view structure fit"),
        )


class Owner(models.Model):
    """A corporation that owns structures"""

    class RotateCharactersType(models.TextChoices):
        """Type of sync to rotate characters for."""

        STRUCTURES = "structures"
        NOTIFICATIONS = "notifications"

        @property
        def last_used_at_name(self) -> str:
            """Name of last used at property."""
            if self is self.STRUCTURES:
                return "structures_last_used_at"
            if self is self.NOTIFICATIONS:
                return "notifications_last_used_at"
            raise NotImplementedError(f"Not defined for: {self}")

        @property
        def esi_cache_duration(self) -> int:
            """ESI cache duration in seconds."""
            if self is self.STRUCTURES:
                return 3600
            if self is self.NOTIFICATIONS:
                return 600
            raise NotImplementedError(f"Not defined for: {self}")

    # PK
    corporation = models.OneToOneField(
        EveCorporationInfo,
        primary_key=True,
        on_delete=models.CASCADE,
        related_name="structure_owner",
        verbose_name=_("corporation"),
        help_text=_("Corporation owning structures"),
    )
    # regular
    are_pocos_public = models.BooleanField(
        default=False,
        verbose_name=_("are pocos public"),
        help_text=_("whether pocos of this owner are shown on public POCO page"),
    )
    assets_last_update_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("assets last update at"),
        help_text=_("when the last successful update happened"),
    )
    character_ownership = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
        help_text="OUTDATED. Has been replaced by OwnerCharacter",  # TODO: Remove
    )
    forwarding_last_update_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("forwarding last update at"),
        help_text=_("When the last successful update happened"),
    )
    has_default_pings_enabled = models.BooleanField(
        default=True,
        verbose_name=_("has default pings enabled"),
        help_text=_(
            "To enable or disable pinging of notifications for this owner "
            "e.g. with @everyone and @here"
        ),
    )
    is_active = models.BooleanField(
        default=True,
        verbose_name=_("is active"),
        help_text=_("Whether this owner is currently included in the sync process"),
    )
    is_alliance_main = models.BooleanField(
        default=False,
        verbose_name=_("is alliance main"),
        help_text=_(
            "Whether alliance wide notifications "
            "are forwarded for this owner (e.g. sov notifications)"
        ),
    )
    is_included_in_service_status = models.BooleanField(
        default=True,
        verbose_name=_("is included in service status"),
        help_text=_(
            "Whether the sync status of this owner is included in "
            "the overall status of this services"
        ),
    )
    is_up = models.BooleanField(
        null=True,
        default=None,
        editable=False,
        verbose_name=_("is up"),
        help_text=_("Whether all services for this owner are currently up"),
    )
    notifications_last_update_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("notifications last update at"),
        help_text=_("When the last successful update happened"),
    )
    ping_groups = models.ManyToManyField(
        Group,
        default=None,
        blank=True,
        related_name="+",
        verbose_name=_("ping groups"),
        help_text=_("Groups to be pinged for each notification. "),
    )
    structures_last_update_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("structures last update at"),
        help_text=_("When the last successful update happened"),
    )
    webhooks = models.ManyToManyField(
        "Webhook",
        default=None,
        blank=True,
        related_name="owners",
        verbose_name=_("webhooks"),
        help_text=_("Notifications are sent to these webhooks."),
    )

    objects = OwnerManager()

    class Meta:
        verbose_name = _("owner")
        verbose_name_plural = _("owners")

    def __str__(self) -> str:
        return str(self.corporation.corporation_name)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(pk={self.pk}, corporation='{self.corporation}')"
        )

    def save(self, *args, **kwargs) -> None:
        if self.is_alliance_main:
            Owner.objects.filter(
                corporation__alliance_id=self.corporation.alliance_id
            ).exclude(corporation__alliance_id__isnull=True).update(
                is_alliance_main=False
            )
            if "update_fields" in kwargs:
                u = set(kwargs["update_fields"])
                u.add("is_alliance_main")
                kwargs["update_fields"] = u
        super().save(*args, **kwargs)

    @property
    def is_structure_sync_fresh(self) -> bool:
        """True if last sync happened with grace time, else False."""
        return bool(
            self.structures_last_update_at
        ) and self.structures_last_update_at > (
            now() - dt.timedelta(minutes=STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES)
        )

    @property
    def is_notification_sync_fresh(self) -> bool:
        """True if last sync happened with grace time, else False."""
        return bool(
            self.notifications_last_update_at
        ) and self.notifications_last_update_at > (
            now() - dt.timedelta(minutes=STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES)
        )

    @property
    def is_forwarding_sync_fresh(self) -> bool:
        """True if last sync happened with grace time, else False."""
        return bool(
            self.forwarding_last_update_at
        ) and self.forwarding_last_update_at > (
            now() - dt.timedelta(minutes=STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES)
        )

    @property
    def is_assets_sync_fresh(self) -> bool:
        """True if last sync happened with grace time, else False."""
        return bool(self.assets_last_update_at) and self.assets_last_update_at > (
            now() - dt.timedelta(minutes=STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES)
        )

    @property
    def are_all_syncs_ok(self) -> bool:
        """returns true if they have been no errors
        and last syncing occurred within alloted time for all sync categories
        """
        return (
            self.is_structure_sync_fresh
            and self.is_notification_sync_fresh
            and self.is_forwarding_sync_fresh
            and self.is_assets_sync_fresh
        )

    def update_is_up(self) -> bool:
        """Check if all services for this owner are up, notify admins if necessary
        and store result.
        """

        def fresh_str(value: bool) -> str:
            return "up" if value else "down"

        is_up = self.are_all_syncs_ok
        if (
            STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED
            and self.is_included_in_service_status
        ):
            if self.is_up and is_up is False:
                title = f"{__title__}: Services are down for {self}"
                msg = (
                    f"Structure services for {self} are down. "
                    "Admin action is likely required to restore services.\n"
                    f"- Structures: {fresh_str(self.is_structure_sync_fresh)}\n"
                    f"- Notifications: {fresh_str(self.is_notification_sync_fresh)}\n"
                    f"- Forwarding: {fresh_str(self.is_forwarding_sync_fresh)}\n"
                    f"- Assets: {fresh_str(self.is_assets_sync_fresh)}\n"
                )
                notify_admins(message=msg, title=title, level="danger")
            elif self.is_up is False and is_up:
                title = f"{__title__}: Services restored for {self}"
                msg = f"Structure services for {self} have been restored. "
                notify_admins(message=msg, title=title, level="success")
            elif self.is_up is None and is_up:
                title = f"{__title__}: Services enabled {self}"
                msg = f"Structure services for {self} have been enabled. "
                notify_admins(message=msg, title=title, level="success")
        if self.is_up != is_up:
            self.is_up = is_up
            self.save(update_fields=["is_up"])
        return is_up

    def add_character(
        self, character_ownership: CharacterOwnership
    ) -> "OwnerCharacter":
        """Add character to this owner.

        Raises ValueError when character does not belong to owner's corporation.
        """
        if (
            character_ownership.character.corporation_id
            != self.corporation.corporation_id
        ):
            raise ValueError(
                f"Character {character_ownership.character} does not belong "
                "to owner corporation."
            )
        obj: OwnerCharacter = self.characters.get_or_create(
            character_ownership=character_ownership
        )[0]
        obj.reset()
        return obj

    def valid_characters_count(self) -> int:
        """Count of valid owner characters."""
        return self.characters.filter(is_enabled=True).count()

    def has_sov(self, eve_solar_system: EveSolarSystem) -> bool:
        """Determine whether this owner has sov in the given solar system."""
        return EveSovereigntyMap.objects.corporation_has_sov(
            eve_solar_system=eve_solar_system, corporation=self.corporation
        )

    def disable_character(
        self,
        character: "OwnerCharacter",
        reason: str,
        max_allowed_errors: int = 0,
    ) -> None:
        """Disable character and notify it's owner and admins about it.

        Args:
        - character: Character to disable
        - reason: User friendly reason for the deletion
        - max_allowed_errors: Maximum number of allowed errors for this type of error
        """
        if character.error_count < max_allowed_errors:
            character.increase_error_count()
            return

        character.disable(reason)

        title = f"{__title__}: {self}: Character has been disabled"
        level = "warning"
        message = (
            f"{character.character_ownership}: {reason}\n"
            "This character caused too many errors and has been disabled. "
            "Administrator action is required to resolve this issue."
        )
        notify(
            user=character.character_ownership.user,
            title=title,
            message=message,
            level=level,
        )
        if not self.valid_characters_count():
            message += (
                " This owner has no configured characters anymore "
                "and it's services are now down."
            )
            level = "danger"

        notify_admins(title=f"FYI: {title}", message=message, level=level)

    def delete_character(self, character: "OwnerCharacter", reason: str) -> None:
        """Delete character and notify it's owner and admin about the reason

        Args:
        - character: Character this error refers to
        - reason: User friendly reason for the deletion
        """
        character.delete()

        title = f"{__title__}: {self}: Invalid character has been removed"
        level = "warning"
        message = (
            f"{character.character_ownership}: {reason}\n"
            "Your character is no longer valid for syncing this owner and has been removed. "
        )
        notify(
            user=character.character_ownership.user,
            title=title,
            message=message,
            level=level,
        )
        if not self.valid_characters_count():
            message += (
                " This owner has no valid characters anymore "
                "and it's services are now down."
            )
            level = "danger"
        notify_admins(title=f"FYI: {title}", message=message, level=level)

    def _rotate_character(
        self,
        character: "OwnerCharacter",
        ignore_schedule: bool,
        rotate_characters: RotateCharactersType,
    ) -> None:
        """Rotate this character such that all are spread evenly
        across the ESI cache duration for each ESI call.
        """
        time_since_last_used = (
            (
                now() - getattr(character, rotate_characters.last_used_at_name)
            ).total_seconds()
            if getattr(character, rotate_characters.last_used_at_name)
            else None
        )
        try:
            minimum_time_between_rotations = max(
                rotate_characters.esi_cache_duration / self.valid_characters_count(),
                60,
            )
        except ZeroDivisionError:
            minimum_time_between_rotations = rotate_characters.esi_cache_duration
        if (
            ignore_schedule
            or not time_since_last_used
            or time_since_last_used >= minimum_time_between_rotations
        ):
            setattr(character, rotate_characters.last_used_at_name, now())
            character.save()

    def fetch_token(
        self,
        rotate_characters: Optional[RotateCharactersType] = None,
        ignore_schedule: bool = False,
    ) -> Token:
        """Fetch a valid token for the owner and return it.

        Args:
            rotate_characters: For which sync type to rotate through characters \
                with every new call
            ignore_schedule: Ignore current schedule when rotating

        Raises TokenError when no valid token can be provided.
        """
        order_by_last_used = (
            rotate_characters.last_used_at_name
            if rotate_characters
            else "notifications_last_used_at"
        )
        enabled_characters: models.QuerySet[OwnerCharacter] = self.characters.filter(
            is_enabled=True
        ).order_by(order_by_last_used)
        for character in enabled_characters:
            if (
                character.character_ownership.character.corporation_id
                != self.corporation.corporation_id
            ):
                corporation_name = self.corporation.corporation_name
                self.delete_character(
                    character=character,
                    reason=f"Character is no longer a member of {corporation_name}",
                )
                continue

            if not character.character_ownership.user.has_perm(
                "structures.add_structure_owner"
            ):
                self.delete_character(
                    character=character,
                    reason="Character no longer has permission to sync",
                )
                continue

            token = character.valid_token()
            if not token:
                self.disable_character(
                    character=character, reason="No valid token found for character"
                )
                continue

            found_character = character
            break  # leave the for loop if we have found a valid token

        else:
            raise TokenError(
                f"{self}: No valid character found for sync. "
                "Service down for this owner."
            )

        if rotate_characters:
            self._rotate_character(
                character=found_character,
                ignore_schedule=ignore_schedule,
                rotate_characters=rotate_characters,
            )
        return token

    def _report_esi_issue(self, action: str, ex: Exception, token: Token):
        """Report an ESI issue to admins."""
        message = f"{self}: Failed to {action} from ESI with token {token} due to {ex}"
        logger.warning(message, exc_info=True)

    def update_structures_esi(self, user: Optional[User] = None):
        """Updates all structures from ESI."""
        token = self.fetch_token(rotate_characters=self.RotateCharactersType.STRUCTURES)
        is_ok = self._fetch_upwell_structures(token)
        if STRUCTURES_FEATURE_CUSTOMS_OFFICES:
            is_ok &= self._fetch_custom_offices(token)
        if STRUCTURES_FEATURE_STARBASES:
            is_ok &= self._fetch_starbases(token)

        if is_ok:
            self.structures_last_update_at = now()
            self.save(update_fields=["structures_last_update_at"])
            if user:
                self._send_report_to_user(
                    topic="structures", topic_count=self.structures.count(), user=user
                )

    def _remove_structures_not_returned_from_esi(
        self, existing_structures: models.QuerySet, new_structures: Iterable[dict]
    ):
        """Remove structures no longer returned from ESI."""
        ids_local = {x.id for x in existing_structures}
        ids_from_esi = {x["structure_id"] for x in new_structures}
        ids_to_remove = ids_local - ids_from_esi
        if len(ids_to_remove) > 0:
            existing_structures.filter(id__in=ids_to_remove).delete()
            logger.info(
                "Removed %d structures which apparently no longer exist.",
                len(ids_to_remove),
            )

    def _fetch_locations_for_assets(
        self, item_ids: Iterable[int], token: Token
    ) -> dict:
        """Fetch locations for given asset items from ESI."""
        item_ids = list(item_ids)
        logger.info(
            "%s: Fetching locations for %d assets from ESI", self, len(item_ids)
        )
        locations_data = []
        for item_ids_chunk in chunks(item_ids, 999):
            try:
                locations_data_chunk = (
                    esi.client.Assets.post_corporations_corporation_id_assets_locations(
                        corporation_id=self.corporation.corporation_id,
                        item_ids=item_ids_chunk,
                        token=token.valid_access_token(),
                    )
                ).results()
            except HTTPNotFound:
                pass
            else:
                locations_data += locations_data_chunk
        positions = {x["item_id"]: x["position"] for x in locations_data}
        return positions

    def _fetch_upwell_structures(self, token: Token) -> bool:
        """Fetch Upwell structures from ESI for self.

        Return True if successful, else False.
        """
        is_ok = True
        # fetch main list of structure for this corporation
        try:
            structures: List[dict] = (
                esi.client.Corporation.get_corporations_corporation_id_structures(
                    corporation_id=self.corporation.corporation_id,
                    token=token.valid_access_token(),
                ).results()
            )
        except OSError as ex:
            self._report_esi_issue("fetch corporation structures", ex, token)
            return False

        # fetch additional information for structures
        if not structures:
            logger.info("%s: No Upwell structures retrieved from ESI", self)
        else:
            logger.info(
                "%s: Fetching additional infos for %d Upwell structures from ESI",
                self,
                len(structures),
            )
            for structure in structures:
                try:
                    structure_info = (
                        esi.client.Universe.get_universe_structures_structure_id(
                            structure_id=structure["structure_id"],
                            token=token.valid_access_token(),
                        )
                    ).results()
                except OSError as ex:
                    if isinstance(ex, HTTPForbidden):
                        logger.error(
                            "Failed to fetch structure with ID #%d belonging to %s, "
                            "because the character '%s' is missing "
                            "docking rights for it.",
                            structure["structure_id"],
                            self,
                            token.character_name,
                        )
                    else:
                        self._report_esi_issue(
                            f"fetch structure #{structure['structure_id']}", ex, token
                        )
                    structure["name"] = "(no data)"
                    is_ok = False
                else:
                    structure["name"] = Structure.extract_name_from_esi_response(
                        structure_info["name"]
                    )
                    structure["position"] = structure_info["position"]

            logger.info(
                "%s: Storing updates for %d upwell structures",
                self,
                len(structures),
            )
            for structure in structures:
                try:
                    Structure.objects.update_or_create_from_dict(structure, self)
                except OSError:
                    logger.warning(
                        "%s: Failed to store update for structure with ID %s",
                        self,
                        structure["structure_id"],
                    )
                    is_ok = False

        if STRUCTURES_DEVELOPER_MODE:
            self._store_raw_data("structures", structures)

        self._remove_structures_not_returned_from_esi(
            existing_structures=self.structures.filter_upwell_structures(),
            new_structures=structures,
        )
        return is_ok

    def _fetch_custom_offices(self, token: Token) -> bool:
        """Fetch custom offices from ESI for this owner.

        Return True when successful, else False.
        """
        structures = {}
        try:
            pocos = esi.client.Planetary_Interaction.get_corporations_corporation_id_customs_offices(
                corporation_id=self.corporation.corporation_id,
                token=token.valid_access_token(),
            ).results()

            if not pocos:
                logger.info("%s: No custom offices retrieved from ESI", self)

            else:
                pocos_2 = {row["office_id"]: row for row in pocos}
                office_ids = list(pocos_2.keys())
                positions = self._fetch_locations_for_assets(office_ids, token)
                names = self._fetch_names_for_pocos(office_ids, token)

                # making sure we have all solar systems loaded
                # incl. their planets for later name matching
                solar_system_ids = {int(obj["system_id"]) for obj in pocos}
                for solar_system_id in solar_system_ids:
                    EveSolarSystem.objects.get_or_create_esi(
                        id=solar_system_id,
                        include_children=True,
                        wait_for_children=True,
                        enabled_sections=[EveSolarSystem.Section.PLANETS],
                    )

                structures.update(
                    self._compile_pocos_for_structures(pocos_2, positions, names)
                )

                self._store_poco_details(structures, pocos_2)

            if STRUCTURES_DEVELOPER_MODE:
                self._store_raw_data("customs_offices", structures)

        except OSError as ex:
            self._report_esi_issue("fetch custom offices", ex, token)
            return False

        self._remove_structures_not_returned_from_esi(
            existing_structures=self.structures.filter_customs_offices(),
            new_structures=structures.values(),
        )
        return True

    def _store_poco_details(self, structures: dict, pocos_2: dict):
        logger.info("%s: Storing updates for %d customs offices", self, len(structures))
        for office_id, structure in structures.items():
            try:
                poco = pocos_2[office_id]
            except KeyError:
                logger.warning(
                    "%s: No details found for this POCO: %d", self, office_id
                )
                continue

            standing_level = PocoDetails.StandingLevel.from_esi(
                poco.get("standing_level")
            )
            structure_obj, _ = Structure.objects.update_or_create_from_dict(
                structure, self
            )
            PocoDetails.objects.update_or_create(
                structure=structure_obj,
                defaults={
                    "alliance_tax_rate": poco.get("alliance_tax_rate"),
                    "allow_access_with_standings": poco.get(
                        "allow_access_with_standings"
                    ),
                    "allow_alliance_access": poco.get("allow_alliance_access"),
                    "bad_standing_tax_rate": poco.get("bad_standing_tax_rate"),
                    "corporation_tax_rate": poco.get("corporation_tax_rate"),
                    "excellent_standing_tax_rate": poco.get(
                        "excellent_standing_tax_rate"
                    ),
                    "good_standing_tax_rate": poco.get("good_standing_tax_rate"),
                    "neutral_standing_tax_rate": poco.get("neutral_standing_tax_rate"),
                    "reinforce_exit_end": poco.get("reinforce_exit_end"),
                    "reinforce_exit_start": poco.get("reinforce_exit_start"),
                    "standing_level": standing_level,
                    "terrible_standing_tax_rate": poco.get(
                        "terrible_standing_tax_rate"
                    ),
                },
            )

    def _compile_pocos_for_structures(self, pocos_2, positions, names) -> dict:
        structures = {}
        for office_id, poco in pocos_2.items():
            planet_name = names.get(office_id, "")
            if planet_name:
                try:
                    eve_planet = EvePlanet.objects.get(name=planet_name)
                except EvePlanet.DoesNotExist:
                    name = ""
                    planet_id = None
                else:
                    planet_id = eve_planet.id
                    name = eve_planet.eve_type.name
            else:
                name = None
                planet_id = None

            reinforce_exit_start = dt.datetime(
                year=2000, month=1, day=1, hour=poco["reinforce_exit_start"]
            )
            reinforce_hour = reinforce_exit_start + dt.timedelta(hours=1)
            structure = {
                "structure_id": office_id,
                "type_id": EveTypeId.CUSTOMS_OFFICE,
                "corporation_id": self.corporation.corporation_id,
                "name": name if name else "",
                "system_id": poco["system_id"],
                "reinforce_hour": reinforce_hour.hour,
                "state": Structure.State.UNKNOWN,
            }
            if planet_id:
                structure["planet_id"] = planet_id

            if office_id in positions:
                structure["position"] = positions[office_id]

            structures[office_id] = structure

        return structures

    def _fetch_names_for_pocos(self, item_ids: list, token: Token) -> dict:
        logger.info(
            "%s: Fetching names for %d custom office names from ESI",
            self,
            len(item_ids),
        )
        names_data = []
        for item_ids_chunk in chunks(item_ids, 999):
            try:
                names_data_chunk = (
                    esi.client.Assets.post_corporations_corporation_id_assets_names(
                        corporation_id=self.corporation.corporation_id,
                        item_ids=item_ids_chunk,
                        token=token.valid_access_token(),
                    )
                ).results()
            except HTTPNotFound:
                pass
            else:
                names_data += names_data_chunk
        names = {x["item_id"]: self._extract_planet_name(x["name"]) for x in names_data}
        return names

    @staticmethod
    def _extract_planet_name(text: str) -> str:
        """Extract name of planet from assert name for a customs office."""
        reg_ex = re.compile(r"Customs Office \((.+)\)")
        matches = reg_ex.match(text)
        return matches.group(1) if matches else ""

    def _fetch_starbases(self, token: Token) -> bool:
        """Fetch starbases from ESI for this owner.

        Return True when successful, else False.
        """
        try:
            character: OwnerCharacter = self.characters.get(
                character_ownership__character__character_id=token.character_id
            )
        except ObjectDoesNotExist:
            return False

        structures = []
        try:
            starbases_data = (
                esi.client.Corporation.get_corporations_corporation_id_starbases(
                    corporation_id=self.corporation.corporation_id,
                    token=token.valid_access_token(),
                )
            ).results()
            if not starbases_data:
                logger.info("%s: This corporation has no starbases.", self)

            else:
                starbases_data = {obj["starbase_id"]: obj for obj in starbases_data}
                names = self._fetch_starbases_names(starbases_data.keys(), token)
                locations = self._fetch_locations_for_assets(
                    starbases_data.keys(), token
                )
                structures += self._convert_starbases_to_structures(
                    starbases_data, names, locations
                )

                self._store_updates_for_starbases(token, structures)

            if STRUCTURES_DEVELOPER_MODE:
                self._store_raw_data("starbases", structures)

        except HTTPForbidden:
            self.disable_character(
                character=character,
                reason=("This character is not a director or CEO"),
                max_allowed_errors=STRUCTURES_ESI_DIRECTOR_ERROR_MAX_RETRIES,
            )
            return False

        except OSError as ex:
            self._report_esi_issue("fetch starbases", ex, token)
            return False

        self._remove_structures_not_returned_from_esi(
            existing_structures=self.structures.filter_starbases(),
            new_structures=structures,
        )
        character.reset_error_counter()
        return True

    def _store_updates_for_starbases(self, token, structures):
        logger.info("%s: Storing updates for %d starbases", self, len(structures))
        for structure in structures:
            structure_obj, _ = Structure.objects.update_or_create_from_dict(
                structure, self
            )
            detail = self._update_starbase_detail(structure=structure_obj, token=token)
            fuel_expires_at = detail.calc_fuel_expires()
            if fuel_expires_at:
                structure_obj.fuel_expires_at = fuel_expires_at
                structure_obj.save()
            if (
                structure_obj.state == Structure.State.POS_REINFORCED
                and structure_obj.state_timer_end
            ):
                GeneratedNotification.objects.get_or_create_from_structure(
                    structure=structure_obj,
                    notif_type=NotificationType.TOWER_REINFORCED_EXTRA,
                )

    def _convert_starbases_to_structures(
        self, starbases_data, names, locations
    ) -> List[dict]:
        structures = []
        for structure_id, starbase in starbases_data.items():
            try:
                name = names[structure_id]
            except KeyError:
                name = "Starbase"
            structure = {
                "structure_id": structure_id,
                "type_id": starbase["type_id"],
                "corporation_id": self.corporation.corporation_id,
                "name": name,
                "system_id": starbase["system_id"],
            }
            if structure_id in locations:
                structure["position"] = locations[structure_id]
            if "state" in starbase:
                structure["state"] = starbase["state"]
            if "moon_id" in starbase:
                structure["moon_id"] = starbase["moon_id"]
            if "fuel_expires" in starbase:
                structure["fuel_expires"] = starbase["fuel_expires"]
            if "reinforced_until" in starbase:
                structure["state_timer_end"] = starbase["reinforced_until"]
            if "unanchors_at" in starbase:
                structure["unanchors_at"] = starbase["unanchors_at"]
            structures.append(structure)

        return structures

    def _fetch_starbases_names(self, item_ids: Iterable, token: Token) -> dict:
        item_ids = list(item_ids)
        logger.info("%s: Fetching names for %d starbases from ESI", self, len(item_ids))
        names_data = []
        for item_ids_chunk in chunks(item_ids, 999):
            try:
                names_data_chunk = (
                    esi.client.Assets.post_corporations_corporation_id_assets_names(
                        corporation_id=self.corporation.corporation_id,
                        item_ids=item_ids_chunk,
                        token=token.valid_access_token(),
                    )
                ).results()
            except HTTPNotFound:
                pass
            else:
                names_data += names_data_chunk
        names = {x["item_id"]: x["name"] for x in names_data}
        return names

    def _update_starbase_detail(self, structure, token: Token) -> StarbaseDetail:
        """Update detail for the starbase from ESI."""
        operation = esi.client.Corporation.get_corporations_corporation_id_starbases_starbase_id(
            corporation_id=structure.owner.corporation.corporation_id,
            starbase_id=structure.id,
            system_id=structure.eve_solar_system.id,
            token=token.valid_access_token(),
        )
        operation.request_config.also_return_response = True
        data, response = operation.results()
        last_modified_at = parsedate_to_datetime(
            response.headers.get("Last-Modified", format_datetime(now()))
        )
        defaults = {
            "allow_alliance_members": data["allow_alliance_members"],
            "allow_corporation_members": data["allow_corporation_members"],
            "anchor_role": StarbaseDetail.Role.from_esi(data["anchor"]),
            "attack_if_at_war": data["attack_if_at_war"],
            "attack_if_other_security_status_dropping": data.get(
                "attack_if_other_security_status_dropping"
            ),
            "attack_security_status_threshold": data.get(
                "attack_security_status_threshold"
            ),
            "attack_standing_threshold": data.get("attack_standing_threshold"),
            "fuel_bay_take_role": StarbaseDetail.Role.from_esi(data["fuel_bay_take"]),
            "fuel_bay_view_role": StarbaseDetail.Role.from_esi(data["fuel_bay_view"]),
            "last_modified_at": last_modified_at,
            "offline_role": StarbaseDetail.Role.from_esi(data["offline"]),
            "online_role": StarbaseDetail.Role.from_esi(data["online"]),
            "unanchor_role": StarbaseDetail.Role.from_esi(data["unanchor"]),
            "use_alliance_standings": data["use_alliance_standings"],
        }
        for fuel in data["fuels"]:
            EveType.objects.get_or_create_esi(id=fuel["type_id"])
        with transaction.atomic():
            detail, _ = StarbaseDetail.objects.update_or_create(
                structure=structure, defaults=defaults
            )
            detail.fuels.all().delete()
            for fuel in data["fuels"]:
                StarbaseDetailFuel.objects.create(
                    eve_type_id=fuel["type_id"],
                    detail=detail,
                    quantity=fuel["quantity"],
                )
        return detail

    def add_or_remove_timers_from_notifications(self):
        """Add/remove timers from esi and generated notification of this owner."""
        self.notification_set.add_or_remove_timers()
        self.generatednotification_set.add_or_remove_timers()

    def fetch_notifications_esi(self, user: Optional[User] = None) -> None:
        """Fetch notifications for this owner from ESI and process them."""
        notifications_count_all = 0
        token = self.fetch_token(
            rotate_characters=self.RotateCharactersType.NOTIFICATIONS
        )
        notifications = self._fetch_notifications_from_esi(token)
        notifications_count_new = self._store_notifications(notifications)
        self._process_moon_notifications()
        if notifications_count_new > 0:
            logger.info(
                "%s: Received %d new notifications from ESI",
                self,
                notifications_count_new,
            )
            notifications_count_all += notifications_count_new
        else:
            logger.info("%s: No new notifications received from ESI", self)

        self.notifications_last_update_at = now()
        self.save(update_fields=["notifications_last_update_at"])
        if user:
            self._send_report_to_user(
                topic="notifications",
                topic_count=notifications_count_all,
                user=user,
            )

    def _fetch_notifications_from_esi(self, token: Token) -> List[dict]:
        """fetching all notifications from ESI for current owner"""
        notifications = esi.client.Character.get_characters_character_id_notifications(
            character_id=token.character_id, token=token.valid_access_token()
        ).results()
        if STRUCTURES_DEVELOPER_MODE:
            self._store_raw_data("notifications", notifications)
        if STRUCTURES_NOTIFICATIONS_ARCHIVING_ENABLED:
            self._store_raw_notifications(notifications)
        logger.debug(
            "%s: Processing %d notifications received from ESI",
            self,
            len(notifications),
        )
        return notifications

    def _store_raw_notifications(self, notifications):
        """Store notifications to disk in continuous file per corp."""
        folder_name = "structures_notifications_archive"
        os.makedirs(folder_name, exist_ok=True)
        corporation_id = self.corporation.corporation_id
        now_str = now().date().isoformat()
        filename = f"{folder_name}/notifications_{corporation_id}_{now_str}.txt"
        logger.info(
            "%s: Storing notifications into archive file: %s", self, format(filename)
        )
        with open(file=filename, mode="a", encoding="utf-8") as file:
            now_str_2 = now().strftime(DATETIME_FORMAT)
            file.write(f"[{now_str_2}] {self.corporation.corporation_ticker}:\n")
            json.dump(
                notifications, file, cls=DjangoJSONEncoder, sort_keys=True, indent=4
            )
            file.write("\n")

    def _store_notifications(self, notifications: List[dict]) -> int:
        """Store new notifications in database.
        Returns number of newly created objects.
        """
        # identify new notifications
        existing_notification_ids = set(
            self.notification_set.values_list("notification_id", flat=True)
        )
        new_notifications = [
            obj
            for obj in notifications
            if obj["notification_id"] not in existing_notification_ids
        ]
        # create new notification objects
        for notification in new_notifications:
            if notification["sender_type"] == "other":
                sender = None
            else:
                sender, _ = EveEntity.objects.get_or_create_esi(
                    id=notification["sender_id"]
                )
            text = notification["text"] if "text" in notification else None
            is_read = notification["is_read"] if "is_read" in notification else None
            # at least one type has a trailing white space
            # which we need to remove
            notif_type = notification["type"].strip()
            Notification.objects.create(
                notification_id=notification["notification_id"],
                owner=self,
                sender=sender,
                timestamp=notification["timestamp"],
                notif_type=notif_type,
                text=text,
                is_read=is_read,
                last_updated=now(),
                created=now(),
            )
        return len(new_notifications)

    def _process_moon_notifications(self):
        """processes notifications for timers if any"""
        empty_refineries = Structure.objects.filter(
            owner=self,
            eve_type__eve_group_id=EveGroupId.REFINERY,
            eve_moon__isnull=True,
        )
        if empty_refineries.exists():
            logger.info(
                "%s: Trying to find moons for up to %d refineries which have no moon.",
                self,
                empty_refineries.count(),
            )
            notifications = (
                self.notification_set.filter(
                    notif_type__in=NotificationType.relevant_for_moonmining()
                )
                .select_related("owner", "sender")
                .order_by("timestamp")
            )
            structure_id_2_moon_id = {}
            for notification in notifications:
                parsed_text = notification.parsed_text()
                moon_id = parsed_text["moonID"]
                structure_id = parsed_text["structureID"]
                structure_id_2_moon_id[structure_id] = moon_id
            for refinery in empty_refineries:
                if refinery.id in structure_id_2_moon_id:
                    logger.info("%s: Updating moon for structure %s", self, refinery)
                    eve_moon, _ = EveMoon.objects.get_or_create_esi(
                        id=structure_id_2_moon_id[refinery.id]
                    )
                    refinery.eve_moon = eve_moon
                    refinery.save()

    def send_new_notifications(self, user: Optional[User] = None):
        """Forward all new notification of this owner to configured webhooks."""
        notifications_count = 0
        cutoff_dt_for_stale = now() - dt.timedelta(
            hours=STRUCTURES_HOURS_UNTIL_STALE_NOTIFICATION
        )
        my_filter = {
            "notif_type__in": (
                Webhook.objects.enabled_notification_types()
                & NotificationType.relevant_for_forwarding()
            ),
            "is_sent": False,
            "timestamp__gte": cutoff_dt_for_stale,
        }
        new_eve_notifications: models.QuerySet[Notification] = (
            self.notification_set.filter(**my_filter)
            .select_related("owner", "sender", "owner__corporation")
            .order_by("timestamp")
        )
        for notif in new_eve_notifications:
            notif.send_to_configured_webhooks()

        new_generated_notifications: models.QuerySet[GeneratedNotification] = (
            self.generatednotification_set.filter(**my_filter)
            .select_related("owner", "owner__corporation")
            .order_by("timestamp")
        )
        for notif in new_generated_notifications:
            notif.send_to_configured_webhooks()

        if (
            not new_eve_notifications.exists()
            and not new_generated_notifications.exists()
        ):
            logger.info("%s: No new notifications found for forwarding", self)

        self.forwarding_last_update_at = now()
        self.save(update_fields=["forwarding_last_update_at"])
        if user:
            self._send_report_to_user(
                topic="notifications", topic_count=notifications_count, user=user
            )

    def _send_report_to_user(self, topic: str, topic_count: int, user: User):
        message_details = "%(count)s %(topic)s synced." % {
            "count": topic_count,
            "topic": topic,
        }
        message = _(
            "Syncing of %(topic)s for %(owner)s %(result)s.\n %(message_details)s"
        ) % {
            "topic": topic,
            "owner": self.corporation.corporation_name,
            "result": _("completed successfully"),
            "message_details": message_details,
        }
        notify(
            user,
            title=_("%(title)s: %(topic)s updated for %(owner)s: %(result)s")
            % {
                "title": _(__title__),
                "topic": topic,
                "owner": self.corporation.corporation_name,
                "result": _("OK"),
            },
            message=message,
            level="success",
        )

    def _store_raw_data(self, name: str, data: Any):
        """store raw data for debug purposes"""
        with open(
            f"{name}_raw_{self.corporation.corporation_id}.json", "w", encoding="utf-8"
        ) as file:
            json.dump(data, file, cls=DjangoJSONEncoder, sort_keys=True, indent=4)

    def update_asset_esi(self, user: Optional[User] = None):
        """Update assets from ESI."""
        token = self.fetch_token()
        assets_data = self._fetch_owner_assets_from_esi(token)
        self._store_items_for_upwell_structures(assets_data)
        self._store_items_for_starbases(assets_data)
        if STRUCTURES_FEATURE_SKYHOOKS:
            self._update_skyhooks_from_assets(assets_data)
            self._resolve_skyhook_planets()
        if user:
            self._send_report_to_user(
                topic="assets", topic_count=self.structures.count(), user=user
            )

    def _fetch_owner_assets_from_esi(self, token: Token) -> dict:
        assets_raw = esi.client.Assets.get_corporations_corporation_id_assets(
            corporation_id=self.corporation.corporation_id,
            token=token.valid_access_token(),
        ).results()
        assets = {asset["item_id"]: asset for asset in assets_raw}
        positions = self._fetch_locations_for_assets(assets.keys(), token)
        for item_id, asset in assets.items():
            asset["position"] = positions[item_id] if item_id in positions else None
        return assets

    def _store_items_for_upwell_structures(self, assets_data: dict):
        structure_ids = set(
            self.structures.filter_upwell_structures().values_list("id", flat=True)
        )
        assets_in_structures = {id: {} for id in structure_ids}
        for item_id, item in assets_data.items():
            location_id = item["location_id"]
            if location_id in structure_ids and item["location_flag"] not in [
                StructureItem.LocationFlag.CORP_DELIVERIES,
                StructureItem.LocationFlag.OFFICE_FOLDER,
                StructureItem.LocationFlag.SECONDARY_STORAGE,
                StructureItem.LocationFlag.AUTOFIT,
            ]:
                assets_in_structures[location_id][item_id] = item
        for structure in self.structures.all():
            structure_items = []
            if structure.id in assets_in_structures.keys():
                structure_assets = assets_in_structures[structure.id]
                has_fitting = [
                    asset
                    for asset in structure_assets.values()
                    if asset["location_flag"]
                    != StructureItem.LocationFlag.QUANTUM_CORE_ROOM
                ]
                has_core = [
                    asset
                    for asset in structure_assets.values()
                    if asset["location_flag"]
                    == StructureItem.LocationFlag.QUANTUM_CORE_ROOM
                ]
                structure.has_fitting = bool(has_fitting)
                structure.has_core = bool(has_core)
                structure.save()

                for asset in structure_assets.values():
                    structure_items.append(
                        StructureItem.from_esi_asset(asset, structure)
                    )

            structure.update_items(structure_items)
            structure.reevaluate_jump_fuel_alerts()

        self.assets_last_update_at = now()
        self.save(update_fields=["assets_last_update_at"])

    def _store_items_for_starbases(self, assets_raw: dict):
        relevant_assets = {
            item_id: item
            for item_id, item in assets_raw.items()
            if item["location_type"] == "solar_system"
            and item["location_flag"] == "AutoFit"
            and item["position"]
        }
        for structure in self.structures.filter_starbases().filter(
            position_x__isnull=False, position_y__isnull=False, position_z__isnull=False
        ):
            structure_items = []
            for item_id, item in relevant_assets.items():
                if (
                    item["location_id"] == structure.eve_solar_system_id
                    and item_id != structure.id
                    and structure.distance_to_object(
                        item["position"]["x"],
                        item["position"]["y"],
                        item["position"]["z"],
                    )
                    < 100_000
                ):
                    structure_items.append(
                        StructureItem.from_esi_asset(item, structure)
                    )
            structure.update_items(structure_items)

    def _update_skyhooks_from_assets(self, assets_data: dict):
        skyhooks = {
            item_id: item
            for item_id, item in assets_data.items()
            if item["type_id"] == EveTypeId.ORBITAL_SKYHOOK
            and item["location_type"] == "solar_system"
            and item["location_flag"] == "AutoFit"
            and item["is_singleton"]
            and item["position"]
        }
        structures = []
        for item in skyhooks.values():
            structures.append(
                {
                    "corporation_id": self.corporation.corporation_id,
                    "type_id": item["type_id"],
                    "position": item["position"],
                    "structure_id": item["item_id"],
                    "system_id": item["location_id"],
                }
            )

        for s in structures:
            Structure.objects.update_or_create_from_dict(s, self)

        self._remove_structures_not_returned_from_esi(
            existing_structures=self.structures.filter_skyhooks(),
            new_structures=structures,
        )

    def _resolve_skyhook_planets(self):
        """Add planets to all unresolved Skyhooks."""
        s: Structure
        for s in self.structures.filter_skyhooks().filter(
            eve_planet__isnull=True,
            position_x__isnull=False,
            position_y__isnull=False,
            position_z__isnull=False,
        ):
            try:
                celestial = s.eve_solar_system.nearest_celestial(
                    x=s.position_x,
                    y=s.position_y,
                    z=s.position_z,
                    group_id=EveGroupId.PLANET,
                )
            except OSError:
                continue

            if not celestial or not isinstance(celestial.eve_object, EvePlanet):
                continue

            s.eve_planet = celestial.eve_object
            s.name = celestial.eve_type.name
            s.save()

    @staticmethod
    def get_esi_scopes() -> List[str]:
        """Return all required ESI scopes."""
        scopes = [
            "esi-corporations.read_structures.v1",
            "esi-universe.read_structures.v1",
            "esi-characters.read_notifications.v1",
            "esi-assets.read_corporation_assets.v1",
        ]
        if STRUCTURES_FEATURE_CUSTOMS_OFFICES:
            scopes += ["esi-planets.read_customs_offices.v1"]
        if STRUCTURES_FEATURE_STARBASES:
            scopes += ["esi-corporations.read_starbases.v1"]
        return scopes


class OwnerCharacter(models.Model):
    """Character for syncing owner data with ESI."""

    owner = models.ForeignKey(
        Owner,
        on_delete=models.CASCADE,
        related_name="characters",
        verbose_name=_("owner"),
    )
    character_ownership = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("character_ownership"),
        help_text="Character used for syncing",
    )
    structures_last_used_at = models.DateTimeField(
        null=True,
        default=None,
        editable=False,
        db_index=True,
        verbose_name=_("structures last used at"),
        help_text=_("When this character was last used for syncing structures"),
    )
    notifications_last_used_at = models.DateTimeField(
        null=True,
        default=None,
        editable=False,
        db_index=True,
        verbose_name=_("notifications last used at"),
        help_text=_("When this character was last used for syncing notifications"),
    )
    error_count = models.PositiveIntegerField(
        default=0,
        editable=False,
        verbose_name=_("error count"),
        help_text="Count of ESI errors which happened with this character.",
    )
    is_enabled = models.BooleanField(
        default=True,
        verbose_name=_("is enabled"),
        help_text=_("Disabled characters are not used for syncing owners"),
    )
    disabled_reason = models.TextField(default="")
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        verbose_name = _("owner character")
        verbose_name_plural = _("owner characters")
        constraints = [
            models.UniqueConstraint(
                fields=["owner", "character_ownership"], name="functional_pk_ownertoken"
            )
        ]

    def __str__(self) -> str:
        return (
            f"{self.owner.corporation.corporation_name}-"
            f"{self.character_ownership.character.character_name}"
        )

    def character_id(self) -> int:
        """Return character ID of this character."""
        return self.character_ownership.character.character_id

    def valid_token(self) -> Optional[Token]:
        """Provide a valid token or None if none can be found."""
        return (
            Token.objects.filter(
                user=self.character_ownership.user,
                character_id=self.character_ownership.character.character_id,
            )
            .require_scopes(Owner.get_esi_scopes())
            .require_valid()
            .first()
        )

    def reset(self) -> None:
        """Resets a disabled owner character."""
        self.is_enabled = True
        self.disabled_reason = ""
        self.error_count = 0
        self.save()

    def reset_error_counter(self) -> None:
        """Reset the error counter"""
        self.error_count = 0
        self.save(update_fields=["error_count"])

    def disable(self, reason: str = "") -> None:
        """Disables a character."""
        self.is_enabled = False
        self.disabled_reason = reason
        self.save()

    def increase_error_count(self):
        """Increase error count of this character by one."""
        with transaction.atomic():
            self.error_count = F("error_count") + 1
            self.save(update_fields=["error_count"])



import datetime as dt
from unittest.mock import patch

from django.utils.timezone import now, utc

from app_utils.esi_testing import EsiClientStub, EsiEndpoint
from app_utils.testing import NoSocketsTestCase

from structures.constants import EveCorporationId
from structures.models import OwnerCharacter, StarbaseDetail, Structure
from structures.tests.testdata.factories import (
    EveEntityCorporationFactory,
    OwnerFactory,
    StarbaseFactory,
    UserMainDefaultOwnerFactory,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.models.owners"


@patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
@patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
@patch(MODULE_PATH + ".esi")
class TestUpdateStarbasesEsi(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user, structures_last_update_at=None)
        cls.corporation_id = cls.owner.corporation.corporation_id
        EveEntityCorporationFactory(
            id=EveCorporationId.DED, name="DED"
        )  # for notifications
        cls.endpoints = [
            EsiEndpoint(
                "Assets",
                "post_corporations_corporation_id_assets_locations",
                "corporation_id",
                needs_token=True,
                data={
                    str(cls.corporation_id): [
                        {
                            "item_id": 1300000000001,
                            "position": {"x": 40.2, "y": 27.3, "z": -19.4},
                        },
                    ]
                },
            ),
            EsiEndpoint(
                "Assets",
                "post_corporations_corporation_id_assets_names",
                "corporation_id",
                needs_token=True,
                data={
                    str(cls.corporation_id): [
                        {"item_id": 1300000000001, "name": "Home Sweat Home"},
                        {"item_id": 1300000000002, "name": "Bat cave"},
                        {"item_id": 1300000000003, "name": "Panic Room"},
                    ]
                },
            ),
            EsiEndpoint(
                "Corporation",
                "get_corporations_corporation_id_starbases",
                "corporation_id",
                needs_token=True,
                data={
                    str(cls.corporation_id): [
                        {
                            "moon_id": 40161465,
                            "starbase_id": 1300000000001,
                            "state": "online",
                            "system_id": 30002537,
                            "type_id": 16213,  # Caldari Control Tower
                            "reinforced_until": dt.datetime(2020, 4, 5, 7, tzinfo=utc),
                        },
                        {
                            "moon_id": 40161466,
                            "starbase_id": 1300000000002,
                            "state": "offline",
                            "system_id": 30002537,
                            "type_id": 20061,  # Caldari Control Tower Medium
                            "unanchors_at": dt.datetime(2020, 5, 5, 7, tzinfo=utc),
                        },
                        {
                            "moon_id": 40029527,
                            "reinforced_until": dt.datetime(2020, 1, 2, 3, tzinfo=utc),
                            "starbase_id": 1300000000003,
                            "state": "reinforced",
                            "system_id": 30000474,
                            "type_id": 20062,  # Caldari Control Tower Small
                        },
                    ]
                },
            ),
            EsiEndpoint(
                "Corporation",
                "get_corporations_corporation_id_starbases_starbase_id",
                ("corporation_id", "starbase_id"),
                needs_token=True,
                data={
                    str(cls.corporation_id): {
                        "1300000000001": {
                            "allow_alliance_members": True,
                            "allow_corporation_members": True,
                            "anchor": "config_starbase_equipment_role",
                            "attack_if_at_war": False,
                            "attack_if_other_security_status_dropping": False,
                            "fuel_bay_take": "config_starbase_equipment_role",
                            "fuel_bay_view": "starbase_fuel_technician_role",
                            "fuels": [
                                {
                                    "quantity": 960,
                                    "type_id": 4051,  # Nitrogen Fuel Block
                                },
                                {
                                    "quantity": 11678,
                                    "type_id": 16275,  # Strontium Clathrates
                                },
                            ],
                            "offline": "config_starbase_equipment_role",
                            "online": "config_starbase_equipment_role",
                            "unanchor": "config_starbase_equipment_role",
                            "use_alliance_standings": True,
                        },
                        "1300000000002": {
                            "allow_alliance_members": True,
                            "allow_corporation_members": True,
                            "anchor": "config_starbase_equipment_role",
                            "attack_if_at_war": False,
                            "attack_if_other_security_status_dropping": False,
                            "fuel_bay_take": "config_starbase_equipment_role",
                            "fuels": [
                                {"quantity": 5, "type_id": 4051},
                                {"quantity": 11678, "type_id": 16275},
                            ],
                            "fuel_bay_view": "starbase_fuel_technician_role",
                            "offline": "config_starbase_equipment_role",
                            "online": "config_starbase_equipment_role",
                            "unanchor": "config_starbase_equipment_role",
                            "use_alliance_standings": True,
                        },
                        "1300000000003": {
                            "allow_alliance_members": True,
                            "allow_corporation_members": True,
                            "anchor": "config_starbase_equipment_role",
                            "attack_if_at_war": False,
                            "attack_if_other_security_status_dropping": False,
                            "fuel_bay_take": "config_starbase_equipment_role",
                            "fuel_bay_view": "starbase_fuel_technician_role",
                            "fuels": [
                                {
                                    "quantity": 1000,
                                    "type_id": 4051,  # Nitrogen Fuel Block
                                },
                                {
                                    "quantity": 11678,
                                    "type_id": 16275,  # Strontium Clathrates
                                },
                            ],
                            "offline": "config_starbase_equipment_role",
                            "online": "config_starbase_equipment_role",
                            "unanchor": "config_starbase_equipment_role",
                            "use_alliance_standings": True,
                        },
                    }
                },
            ),
            EsiEndpoint(
                "Corporation",
                "get_corporations_corporation_id_structures",
                "corporation_id",
                needs_token=True,
                data={str(cls.corporation_id): [], "2005": []},
            ),  # TODO: Remove once possible
        ]
        cls.esi_client_stub = EsiClientStub.create_from_endpoints(cls.endpoints)

    def test_can_sync_starbases(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)

        # must contain all expected structures
        expected = {1300000000001, 1300000000002, 1300000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

        # verify attributes for POS
        structure = Structure.objects.get(id=1300000000001)
        self.assertEqual(structure.name, "Home Sweat Home")
        self.assertEqual(structure.eve_solar_system_id, 30002537)
        self.assertEqual(
            int(structure.owner.corporation.corporation_id), self.corporation_id
        )
        self.assertEqual(structure.eve_type_id, 16213)
        self.assertEqual(structure.state, Structure.State.POS_ONLINE)
        self.assertEqual(structure.eve_moon_id, 40161465)
        self.assertEqual(
            structure.state_timer_end, dt.datetime(2020, 4, 5, 7, 0, 0, tzinfo=utc)
        )
        self.assertAlmostEqual(
            structure.fuel_expires_at,
            now() + dt.timedelta(hours=24),
            delta=dt.timedelta(seconds=30),
        )
        self.assertEqual(structure.position_x, 40.2)
        self.assertEqual(structure.position_y, 27.3)
        self.assertEqual(structure.position_z, -19.4)
        # verify details
        detail = structure.starbase_detail
        self.assertTrue(detail.allow_alliance_members)
        self.assertTrue(detail.allow_corporation_members)
        self.assertEqual(
            detail.anchor_role, StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE
        )
        self.assertFalse(detail.attack_if_at_war)
        self.assertFalse(detail.attack_if_other_security_status_dropping)
        self.assertEqual(
            detail.anchor_role, StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE
        )
        self.assertEqual(
            detail.fuel_bay_take_role,
            StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        )
        self.assertEqual(
            detail.fuel_bay_view_role,
            StarbaseDetail.Role.STARBASE_FUEL_TECHNICIAN_ROLE,
        )
        self.assertEqual(
            detail.offline_role,
            StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        )
        self.assertEqual(
            detail.online_role,
            StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        )
        self.assertEqual(
            detail.unanchor_role,
            StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        )
        self.assertTrue(detail.use_alliance_standings)
        # fuels
        self.assertEqual(detail.fuels.count(), 2)
        self.assertEqual(detail.fuels.get(eve_type_id=4051).quantity, 960)
        self.assertEqual(detail.fuels.get(eve_type_id=16275).quantity, 11678)

        structure = Structure.objects.get(id=1300000000002)
        self.assertEqual(structure.name, "Bat cave")
        self.assertEqual(structure.eve_solar_system_id, 30002537)
        self.assertEqual(
            int(structure.owner.corporation.corporation_id), self.corporation_id
        )
        self.assertEqual(structure.eve_type_id, 20061)
        self.assertEqual(structure.state, Structure.State.POS_OFFLINE)
        self.assertEqual(structure.eve_moon_id, 40161466)
        self.assertEqual(
            structure.unanchors_at, dt.datetime(2020, 5, 5, 7, 0, 0, tzinfo=utc)
        )
        self.assertIsNone(structure.fuel_expires_at)
        self.assertFalse(structure.generatednotification_set.exists())

        structure = Structure.objects.get(id=1300000000003)
        self.assertEqual(structure.name, "Panic Room")
        self.assertEqual(structure.eve_solar_system_id, 30000474)
        self.assertEqual(
            int(structure.owner.corporation.corporation_id), self.corporation_id
        )
        self.assertEqual(structure.eve_type_id, 20062)
        self.assertEqual(structure.state, Structure.State.POS_REINFORCED)
        self.assertEqual(structure.eve_moon_id, 40029527)
        self.assertAlmostEqual(
            structure.fuel_expires_at,
            now() + dt.timedelta(seconds=360_000),
            delta=dt.timedelta(seconds=30),
        )
        self.assertEqual(
            structure.state_timer_end, dt.datetime(2020, 1, 2, 3, tzinfo=utc)
        )
        self.assertTrue(structure.generatednotification_set.exists())

    # @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    # @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    # @patch(MODULE_PATH + ".notify", spec=True)
    # def test_can_sync_all_structures_and_notify_user(self, mock_notify, mock_esi):
    #     # given
    #     mock_esi.client = self.esi_client_stub
    #     owner = OwnerFactory(user=self.user, structures_last_update_at=None)

    #     # when
    #     owner.update_structures_esi(user=self.user)

    #     # then
    #     owner.refresh_from_db()
    #     self.assertTrue(owner.is_structure_sync_fresh)

    #     # must contain all expected structures
    #     expected = {
    #         1200000000003,
    #         1200000000004,
    #         1200000000005,
    #         1200000000006,
    #         1200000000099,
    #         1300000000001,
    #         1300000000002,
    #         1300000000003,
    #     }
    #     self.assertSetEqual(owner.structures.ids(), expected)

    #     # user report has been sent
    #     self.assertTrue(mock_notify.called)

    def test_should_not_break_on_http_error_when_fetching_starbases(self, mock_esi):
        # given
        new_endpoint = EsiEndpoint(
            "Corporation",
            "get_corporations_corporation_id_starbases",
            http_error_code=500,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        expected = set()
        self.assertSetEqual(owner.structures.ids(), expected)

    @patch(MODULE_PATH + ".STRUCTURES_ESI_DIRECTOR_ERROR_MAX_RETRIES", 3)
    @patch(MODULE_PATH + ".notify", spec=True)
    def test_should_mark_error_when_character_not_director_while_updating_starbases(
        self, mock_notify, mock_esi
    ):
        # given
        new_endpoint = EsiEndpoint(
            "Corporation",
            "get_corporations_corporation_id_starbases",
            http_error_code=403,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        self.assertTrue(mock_notify)
        character: OwnerCharacter = owner.characters.first()
        self.assertEqual(character.error_count, 1)
        self.assertTrue(character.is_enabled)

    @patch(MODULE_PATH + ".STRUCTURES_ESI_DIRECTOR_ERROR_MAX_RETRIES", 3)
    @patch(MODULE_PATH + ".notify", spec=True)
    def test_should_disable_character_when_not_director_while_updating_starbases(
        self, mock_notify, mock_esi
    ):
        # given
        new_endpoint = EsiEndpoint(
            "Corporation",
            "get_corporations_corporation_id_starbases",
            http_error_code=403,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        character: OwnerCharacter = owner.characters.first()
        character.error_count = 3
        character.save()
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        self.assertTrue(mock_notify)
        character.refresh_from_db()
        self.assertFalse(character.is_enabled)

    def test_should_reset_error_count_for_character_when_successful(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        character: OwnerCharacter = owner.characters.first()
        character.error_count = 3
        character.save()
        # when
        owner.update_structures_esi()
        # then
        character.refresh_from_db()
        self.assertTrue(character.is_enabled)
        self.assertEqual(character.error_count, 0)

    def test_should_remove_old_starbases(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        StarbaseFactory(owner=owner, id=1300000000099, name="delete-me")
        # when
        owner.update_structures_esi()
        # then
        expected = {1300000000001, 1300000000002, 1300000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

    def test_should_not_delete_existing_starbases_when_update_failed(self, mock_esi):
        # given

        new_endpoint = EsiEndpoint(
            "Corporation",
            "get_corporations_corporation_id_starbases",
            http_error_code=500,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        StarbaseFactory(owner=owner, id=1300000000001)
        StarbaseFactory(owner=owner, id=1300000000002)
        # when
        owner.update_structures_esi()
        # then
        # self.assertFalse(owner.is_structure_sync_fresh)
        expected = {1300000000001, 1300000000002}
        self.assertSetEqual(owner.structures.ids(), expected)

    def test_should_not_break_when_starbase_names_not_found(self, mock_esi):
        # given
        new_endpoint = EsiEndpoint(
            "Assets",
            "post_corporations_corporation_id_assets_names",
            http_error_code=404,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        expected = {1300000000001, 1300000000002, 1300000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

    # @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    # @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    # def test_should_notify_admins_when_service_is_restored(
    #     self, mock_esi_client
    # ):
    #     # given
    #     mock_esi_client.side_effect = esi_mock_client
    #     owner = OwnerFactory(user=self.user, structures_last_update_at=None)
    #     owner.is_structure_sync_fresh = False
    #     owner.save()
    #     # when
    #     owner.update_structures_esi()
    #     # then
    #     owner.refresh_from_db()
    #     self.assertTrue(owner.is_structure_sync_fresh)



import datetime as dt

from django.test import TestCase
from django.utils.safestring import mark_safe
from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveType
from eveuniverse.tests.testdata.factories_2 import EveTypeFactory

from structures.helpers import (
    datetime_almost_equal,
    floating_icon_with_text_html,
    get_or_create_esi_obj,
    get_or_create_eve_entity,
    get_or_create_eve_type,
    hours_until_deadline,
    is_absolute_url,
)

ICON_URL = "https://images.evetech.net/types/670/icon?size=64"


class TestDatetimeAlmostEqual(TestCase):
    def test_should_return_true(self):
        # given
        d1 = now() + dt.timedelta(hours=0, minutes=55)
        d2 = now() + dt.timedelta(hours=1, minutes=5)
        # when / then
        self.assertTrue(datetime_almost_equal(d1, d2, 3600))
        self.assertTrue(datetime_almost_equal(d2, d1, 3600))
        self.assertFalse(datetime_almost_equal(d2, d1, 60))
        self.assertFalse(datetime_almost_equal(d1, d2, 60))

    def test_should_return_false(self):
        # given
        d1 = now() + dt.timedelta(hours=0, minutes=55)
        d2 = now() + dt.timedelta(hours=1, minutes=5)
        # when / then
        self.assertFalse(datetime_almost_equal(d2, d1, 60))
        self.assertFalse(datetime_almost_equal(d1, d2, 60))

    def test_should_return_false_for_none_dates(self):
        # given
        d1 = now() + dt.timedelta(hours=0, minutes=55)
        # when / then
        self.assertFalse(datetime_almost_equal(d1, None, 3600))
        self.assertFalse(datetime_almost_equal(None, d1, 3600))


class TestHoursUntilDeadline(TestCase):
    def test_should_return_correct_value_for_two_datetimes(self):
        # given
        d1 = now()
        d2 = d1 - dt.timedelta(hours=3)
        # when / then
        self.assertEqual(hours_until_deadline(d1, d2), 3)

    def test_should_return_correct_value_for_one_datetimes(self):
        # given
        d1 = now() + dt.timedelta(hours=3)
        # when / then
        self.assertAlmostEqual(hours_until_deadline(d1), 3, delta=0.1)

    def test_should_raise_error_when_deadline_is_not_a_datetime(self):
        with self.assertRaises(TypeError):
            hours_until_deadline(None)


class TestIsAbsoluteUrl(TestCase):
    def test_should_detect_absolute_urls(self):
        cases = [
            ("https://www.google.com", True),
            ("http://www.google.com", True),
            ("www.google.com", False),
            ("", False),
            ("/abc/x", False),
            (None, False),
        ]
        for url, expected_result in cases:
            with self.subTest(url=url):
                self.assertIs(is_absolute_url(url), expected_result)


class TestGetOrCreateObjs(TestCase):
    def test_should_return_existing_obj_generic(self):
        # given
        obj = EveTypeFactory()
        # when
        obj_2 = get_or_create_esi_obj(EveType, id=obj.id)
        # then
        self.assertEqual(obj, obj_2)

    def test_should_return_existing_obj_type(self):
        # given
        obj = EveTypeFactory()
        # when
        obj_2 = get_or_create_eve_type(id=obj.id)
        # then
        self.assertEqual(obj, obj_2)

    def test_should_return_existing_obj_entity(self):
        # given
        obj = EveEntity.objects.create(
            id=99, name="test", category=EveEntity.CATEGORY_CHARACTER
        )
        # when
        obj_2 = get_or_create_eve_entity(id=obj.id)
        # then
        self.assertEqual(obj, obj_2)


class TestIconWithParagraphHtml(TestCase):
    def test_should_create_html_with_one_line(self):
        # when
        result = floating_icon_with_text_html(ICON_URL, ["Alpha"])
        # then
        expected = (
            '<p><img src="https://images.evetech.net/types/670/icon?size=64" '
            'class="floating-icon">Alpha</p>'
        )
        self.assertEqual(result, expected)

    def test_should_create_html_with_two_line(self):
        # when
        result = floating_icon_with_text_html("#", ["Alpha", "Bravo"])
        expected = '<p><img src="#" class="floating-icon">Alpha<br>Bravo</p>'
        self.assertEqual(result, expected)

    def test_should_create_html_and_detect_safe_strings(self):
        # when
        result = floating_icon_with_text_html(
            "#", [mark_safe('<a href="#">Alpha</a>'), "Bravo"]
        )
        expected = (
            '<p><img src="#" class="floating-icon"><a href="#">Alpha</a><br>Bravo</p>'
        )
        self.assertEqual(result, expected)

    def test_should_create_html_and_escape_unsafe_strings(self):
        # when
        result = floating_icon_with_text_html("#", ['<a href="#">Alpha</a>', "Bravo"])
        expected = (
            '<p><img src="#" class="floating-icon">'
            "&lt;a href=&quot;#&quot;&gt;Alpha&lt;/a&gt;<br>Bravo</p>"
        )
        self.assertEqual(result, expected)



from app_utils.testing import NoSocketsTestCase

from structures.tests.testdata.factories import OwnerCharacterFactory, OwnerFactory

MODULE_PATH = "structures.models.owners"


class TestOwnerCharacter(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.owner = OwnerFactory()

    def test_should_return_str(self):
        # given
        character = OwnerCharacterFactory(owner=self.owner)
        # when/then
        self.assertTrue(str(character))

    def test_can_reset_character(self):
        # given
        character = OwnerCharacterFactory(
            owner=self.owner, is_enabled=False, disabled_reason="reason", error_count=42
        )
        # when
        character.reset()
        # then
        character.refresh_from_db()
        self.assertTrue(character.is_enabled)
        self.assertFalse(character.disabled_reason)
        self.assertFalse(character.error_count)



"""Skyhook embeds."""

# pylint: disable=missing-class-docstring

import dhooks_lite

from django.utils.translation import gettext as _

from app_utils.datetime import ldap_time_2_datetime

from structures.models import Notification, Webhook

from .helpers import (
    gen_corporation_link,
    gen_solar_system_text,
    target_datetime_formatted,
)
from .main import NotificationBaseEmbed


class NotificationSkyhookEmbed(NotificationBaseEmbed):
    """Base class for most structure related notification embeds."""

    def __init__(self, n: Notification) -> None:
        super().__init__(n)
        type_ = n.eve_structure_type("typeID")
        solar_system = n.eve_solar_system("solarsystemID")
        self._description = _(
            "The %(structure_type)s at **%(planet)s** "
            "in %(solar_system)s belonging to %(owner_link)s "
        ) % {
            "structure_type": type_,
            "planet": n.eve_planet(),
            "solar_system": gen_solar_system_text(solar_system),
            "owner_link": gen_corporation_link(str(n.owner)),
        }
        self._thumbnail = dhooks_lite.Thumbnail(
            type_.icon_url(size=self.ICON_DEFAULT_SIZE)
        )


class NotificationSkyhookDeployed(NotificationSkyhookEmbed):
    def __init__(self, n: Notification) -> None:
        super().__init__(n)
        self._title = _("Skyhook has started onlining")
        self._description += _("has started onlining.")
        self._color = Webhook.Color.INFO


class NotificationSkyhookDestroyed(NotificationSkyhookEmbed):
    def __init__(self, n: Notification) -> None:
        super().__init__(n)
        self._title = _("Skyhook destroyed")
        self._description += _("has been destroyed.")
        self._color = Webhook.Color.DANGER


class NotificationSkyhookLostShield(NotificationSkyhookEmbed):
    def __init__(self, n: Notification) -> None:
        super().__init__(n)
        self._title = _("Skyhook lost shield")
        timer_ends_at = ldap_time_2_datetime(self._data["timestamp"])
        self._description += _(
            "has lost its shields and is now in reinforcement state until : %s"
        ) % target_datetime_formatted(timer_ends_at)
        self._color = Webhook.Color.DANGER


class NotificationSkyhookOnline(NotificationSkyhookEmbed):
    def __init__(self, n: Notification) -> None:
        super().__init__(n)
        self._title = _("Skyhook online")
        self._description += _("is now online.")
        self._color = Webhook.Color.SUCCESS


class NotificationSkyhookUnderAttack(NotificationSkyhookEmbed):
    def __init__(self, n: Notification) -> None:
        super().__init__(n)
        self._title = _("Skyhook under attack")
        self._description += _(
            "is under attack by %(attacker)s.\n"
            "%(damage_text)s"
            % {
                "attacker": self.gen_attacker_link(),
                "damage_text": self.compile_damage_text("Percentage"),
            }
        )
        self._color = Webhook.Color.DANGER



import datetime as dt
from unittest.mock import patch

from django.contrib import admin
from django.contrib.admin.sites import AdminSite
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils.timezone import now

from structures.admin import (
    NotificationAdmin,
    OwnerAdmin,
    OwnerAllianceFilter,
    OwnerCorporationsFilter,
    RenderableNotificationFilter,
    StructureAdmin,
    StructureFuelAlertConfigAdmin,
    WebhookAdmin,
)
from structures.core.notification_types import NotificationType
from structures.models import (
    FuelAlertConfig,
    Notification,
    Owner,
    Structure,
    StructureTag,
    Webhook,
)

from .testdata.factories import (
    EveAllianceInfoFactory,
    EveCorporationInfoFactory,
    FuelAlertConfigFactory,
    NotificationFactory,
    OwnerCharacterFactory,
    OwnerFactory,
    PocoFactory,
    StarbaseFactory,
    StructureFactory,
    StructureTagFactory,
    SuperuserFactory,
    WebhookFactory,
)
from .testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.admin"


class MockRequest(object):
    def __init__(self, user=None):
        self.user = user


class TestFuelNotificationConfigAdminView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.defaults = {
            "is_enabled": True,
            "channel_ping_type": Webhook.PingType.HERE,
            "color": Webhook.Color.WARNING,
        }
        cls.user = SuperuserFactory()
        load_eveuniverse()

    def test_should_create_new_config(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 12, "end": 5, "repeat": 2}},
        )
        # then
        self.assertRedirects(
            response, reverse("admin:structures_fuelalertconfig_changelist")
        )
        self.assertEqual(FuelAlertConfig.objects.count(), 1)

    def test_should_update_existing_config(self):
        # given
        self.client.force_login(self.user)
        config = FuelAlertConfigFactory(start=48, end=24, repeat=12)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_change", args=[config.pk]),
            data={**self.defaults, **{"start": 48, "end": 0, "repeat": 2}},
        )
        # then
        self.assertRedirects(
            response, reverse("admin:structures_fuelalertconfig_changelist")
        )
        self.assertEqual(FuelAlertConfig.objects.count(), 1)

    def test_should_remove_existing_fuel_notifications_when_timing_changed(self):
        # given
        self.client.force_login(self.user)
        config = FuelAlertConfigFactory(start=48, end=24, repeat=12)
        structure = StructureFactory()
        structure.structure_fuel_alerts.create(
            config=config, structure=structure, hours=5
        )
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_change", args=[config.pk]),
            data={**self.defaults, **{"start": 48, "end": 0, "repeat": 2}},
        )
        # then
        self.assertRedirects(
            response, reverse("admin:structures_fuelalertconfig_changelist")
        )
        self.assertEqual(structure.structure_fuel_alerts.count(), 0)

    def test_should_not_remove_existing_fuel_notifications_on_other_changes(self):
        # given
        self.client.force_login(self.user)
        config = FuelAlertConfigFactory(start=48, end=24, repeat=12)
        structure = StructureFactory()
        structure.structure_fuel_alerts.create(
            config=config, structure=structure, hours=5
        )
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_change", args=[config.pk]),
            data={
                **self.defaults,
                **{"start": 48, "end": 24, "repeat": 12, "is_enabled": False},
            },
        )
        # then
        self.assertRedirects(
            response, reverse("admin:structures_fuelalertconfig_changelist")
        )
        self.assertEqual(structure.structure_fuel_alerts.count(), 1)

    def test_should_not_allow_end_before_start(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 1, "end": 2, "repeat": 1}},
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "errornote")
        self.assertEqual(FuelAlertConfig.objects.count(), 0)

    def test_should_not_allow_invalid_frequency(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 48, "end": 24, "repeat": 36}},
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "errornote")
        self.assertEqual(FuelAlertConfig.objects.count(), 0)

    def test_should_not_allow_creating_overlapping(self):
        # given
        self.client.force_login(self.user)
        FuelAlertConfigFactory(start=48, end=24, repeat=12)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 36, "end": 0, "repeat": 8}},
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "errornote")
        self.assertEqual(FuelAlertConfig.objects.count(), 1)

    def test_should_work_with_empty_end_field(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 36, "repeat": 8}},
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "errornote")
        self.assertEqual(FuelAlertConfig.objects.count(), 0)

    def test_should_work_with_empty_start_field(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 36, "end": 8}},
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "errornote")
        self.assertEqual(FuelAlertConfig.objects.count(), 0)

    def test_should_work_with_empty_repeat_field(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 36, "end": 8}},
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "errornote")
        self.assertEqual(FuelAlertConfig.objects.count(), 0)


class TestStructureFuelAlertAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.modeladmin = StructureFuelAlertConfigAdmin(
            model=FuelAlertConfig, admin_site=AdminSite()
        )
        load_eveuniverse()

    @patch(MODULE_PATH + ".StructureFuelAlertConfigAdmin.message_user", spec=True)
    @patch(MODULE_PATH + ".tasks", spec=True)
    def test_should_send_fuel_notifications(self, mock_tasks, mock_message_user):
        # given
        config = FuelAlertConfigFactory()
        request = MockRequest()
        queryset = FuelAlertConfig.objects.filter(pk=config.pk)
        # when
        self.modeladmin.send_fuel_notifications(request, queryset)
        # then
        self.assertTrue(mock_tasks.send_queued_messages_for_webhooks.called)
        self.assertTrue(mock_message_user.called)


class TestNotificationAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        cls.modeladmin = NotificationAdmin(model=Notification, admin_site=AdminSite())
        cls.user = SuperuserFactory()
        cls.owner = OwnerFactory()

    def test_structures_when_structure_related(self):
        # given
        obj = NotificationFactory(
            owner=self.owner, notif_type=NotificationType.STRUCTURE_LOST_SHIELD
        )
        # when
        result = self.modeladmin._structures(obj)
        self.assertEqual(result, "?")

    def test_structures_when_not_structure_related(self):
        # given
        obj = NotificationFactory(
            owner=self.owner, notif_type=NotificationType.CHAR_APP_ACCEPT_MSG
        )
        # when
        result = self.modeladmin._structures(obj)
        # then
        self.assertIsNone(result)

    # FIXME: Does not seam to work with special prefetch list
    # def test_webhooks(self):
    #     self.owner.webhooks.add(Webhook.objects.get(name="Test Webhook 2"))
    #     self.assertEqual(
    #         self.modeladmin._webhooks(self.obj), "Test Webhook 1, Test Webhook 2"
    #     )

    @patch(MODULE_PATH + ".NotificationAdmin.message_user", spec=True)
    def test_action_mark_as_sent(self, mock_message_user):
        # given
        notif = NotificationFactory(owner=self.owner, is_sent=False)
        queryset = Notification.objects.all()
        # when
        self.modeladmin.mark_as_sent(MockRequest(self.user), queryset)
        # then
        notif.refresh_from_db()
        self.assertTrue(notif.is_sent)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".NotificationAdmin.message_user", spec=True)
    def test_action_mark_as_unsent(self, mock_message_user):
        # given
        notif = NotificationFactory(owner=self.owner, is_sent=True)
        queryset = Notification.objects.all()
        # when
        self.modeladmin.mark_as_unsent(MockRequest(self.user), queryset)
        # then
        notif.refresh_from_db()
        self.assertFalse(notif.is_sent)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".NotificationAdmin.message_user", spec=True)
    @patch(MODULE_PATH + ".tasks.send_queued_messages_for_webhooks")
    def test_action_send_to_webhook(self, mock_task, mock_message_user):
        # given
        NotificationFactory(
            owner=self.owner, notif_type=NotificationType.STRUCTURE_LOST_SHIELD
        )
        queryset = Notification.objects.all()
        # when
        with patch(MODULE_PATH + ".Notification.send_to_webhook", return_value=True):
            self.modeladmin.send_to_configured_webhooks(
                MockRequest(self.user), queryset
            )
        # then
        self.assertEqual(mock_task.call_count, 1)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".NotificationAdmin.message_user", spec=True)
    @patch(MODULE_PATH + ".Notification.add_or_remove_timer")
    def test_action_process_for_timerboard(
        self, mock_process_for_timerboard, mock_message_user
    ):
        # given
        NotificationFactory(
            owner=self.owner, notif_type=NotificationType.STRUCTURE_LOST_SHIELD
        )
        queryset = Notification.objects.all()
        # when
        self.modeladmin.add_or_remove_timer(MockRequest(self.user), queryset)
        # then
        self.assertEqual(mock_process_for_timerboard.call_count, 1)
        self.assertTrue(mock_message_user.called)

    def test_filter_renderable_notifications(self):
        class NotificationAdminTest(admin.ModelAdmin):
            list_filter = (RenderableNotificationFilter,)

        # create test data
        positive_notif = NotificationFactory(
            owner=self.owner,
            notif_type=NotificationType.WAR_CORPORATION_BECAME_ELIGIBLE,
        )
        NotificationFactory(owner=self.owner, notif_type="unknown")
        modeladmin = NotificationAdminTest(Notification, AdminSite())

        # Make sure the lookups are correct
        request = self.factory.get("/")
        request.user = self.user
        changelist = modeladmin.get_changelist_instance(request)
        filterspec = changelist.get_filters(request)[0][0]
        expected = [("yes", "yes"), ("no", "no")]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned
        request = self.factory.get("/", {"notification_renderable": "yes"})
        request.user = self.user
        changelist = modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = Notification.objects.filter(pk=positive_notif.pk)
        self.assertSetEqual(set(queryset), set(expected))


class TestNotificationAdminWebhooks(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        cls.modeladmin = NotificationAdmin(model=Notification, admin_site=AdminSite())
        cls.user = SuperuserFactory()

    def test_should_return_name_of_owner_webhook(self):
        # given
        owner = OwnerFactory(webhooks__name="Alpha")
        obj = NotificationFactory(
            owner=owner, notif_type=NotificationType.STRUCTURE_LOST_SHIELD
        )
        obj2 = self.modeladmin.get_queryset(MockRequest(user=self.user)).get(pk=obj.pk)
        # when
        result = self.modeladmin._webhooks(obj2)
        # then
        self.assertEqual("Alpha", result)

    def test_should_report_missing_webhook(self):
        # given
        owner = OwnerFactory(webhooks=False)
        obj = NotificationFactory(
            owner=owner, notif_type=NotificationType.STRUCTURE_LOST_SHIELD
        )
        obj2 = self.modeladmin.get_queryset(MockRequest(user=self.user)).get(pk=obj.pk)
        # when
        result = self.modeladmin._webhooks(obj2)
        # then
        self.assertIn("Not configured", result)

    def test_should_report_when_webhooks_not_configured_for_this_notif_type(self):
        # given
        owner = OwnerFactory()
        obj = NotificationFactory(
            owner=owner, notif_type=NotificationType.SOV_ENTOSIS_CAPTURE_STARTED
        )
        obj2 = self.modeladmin.get_queryset(MockRequest(user=self.user)).get(pk=obj.pk)
        # when
        result = self.modeladmin._webhooks(obj2)
        # then
        self.assertIn("Not configured", result)


class TestOwnerAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.modeladmin = OwnerAdmin(model=Owner, admin_site=AdminSite())
        load_eveuniverse()
        cls.user = SuperuserFactory()
        cls.alliance = EveAllianceInfoFactory(
            alliance_id=3001, alliance_name="Wayne Enterprises"
        )
        cls.corporation = EveCorporationInfoFactory(
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            alliance=cls.alliance,
        )

    def test_corporation(self):
        obj = OwnerFactory(corporation=self.corporation)
        self.assertEqual(self.modeladmin._corporation(obj), "Wayne Technologies")

    def test_alliance_normal(self):
        obj = OwnerFactory(corporation=self.corporation)
        self.assertEqual(self.modeladmin._alliance(obj), "Wayne Enterprises")

    def test_alliance_none(self):
        corporation = EveCorporationInfoFactory(create_alliance=False)
        obj = OwnerFactory(corporation=corporation)
        self.assertIsNone(self.modeladmin._alliance(obj))

    def test_webhooks(self):
        # given
        webhook_1 = WebhookFactory(name="Test Webhook 1")
        webhook_2 = WebhookFactory(name="Test Webhook 2")
        obj = OwnerFactory(webhooks=[webhook_1, webhook_2])
        # when/then
        self.assertEqual(
            self.modeladmin._webhooks(obj), "Test Webhook 1<br>Test Webhook 2"
        )

    @patch(MODULE_PATH + ".OwnerAdmin.message_user", spec=True)
    @patch(MODULE_PATH + ".tasks.update_structures_for_owner")
    def test_action_update_structures(self, mock_task, mock_message_user):
        # given
        OwnerFactory()
        queryset = Owner.objects.all()
        # when
        self.modeladmin.update_structures(MockRequest(self.user), queryset)
        # then
        self.assertEqual(mock_task.delay.call_count, 1)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".OwnerAdmin.message_user", spec=True)
    @patch(MODULE_PATH + ".tasks.process_notifications_for_owner")
    def test_action_fetch_notifications(self, mock_task, mock_message_user):
        # given
        OwnerFactory()
        queryset = Owner.objects.all()
        # when
        self.modeladmin.fetch_notifications(MockRequest(self.user), queryset)
        # then
        self.assertEqual(mock_task.delay.call_count, 1)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".OwnerAdmin.message_user", spec=True)
    def test_action_reset_characters(self, mock_message_user):
        # given
        owner_1 = OwnerFactory(characters=False)
        character_1 = OwnerCharacterFactory(owner=owner_1, is_enabled=False)
        owner_2 = OwnerFactory(characters=False)
        character_2 = OwnerCharacterFactory(owner=owner_2, is_enabled=False)
        # when
        queryset = Owner.objects.filter(pk=owner_1.pk)
        self.modeladmin.reset_characters(MockRequest(self.user), queryset)
        # then
        self.assertTrue(mock_message_user.called)
        character_1.refresh_from_db()
        self.assertTrue(character_1.is_enabled)
        character_2.refresh_from_db()
        self.assertFalse(character_2.is_enabled)

    def test_should_return_empty_turnaround_times(self):
        # given
        obj = OwnerFactory()
        # when
        result = self.modeladmin._avg_turnaround_time(obj)
        # then
        self.assertEqual(result, "- | - | -")

    @patch(MODULE_PATH + ".app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_SHORT", 5)
    @patch(MODULE_PATH + ".app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_MEDIUM", 15)
    @patch(MODULE_PATH + ".app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_LONG", 50)
    @patch(
        MODULE_PATH + ".app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_MAX_VALID", 3600
    )
    def test_should_return_correct_turnaround_times(self):
        # given
        my_owner = OwnerFactory()
        my_now = now()
        NotificationFactory(
            owner=my_owner,
            timestamp=my_now,
            created=my_now + dt.timedelta(seconds=3601),
        )
        for i in range(50):
            timestamp = my_now + dt.timedelta(minutes=i)
            NotificationFactory(
                owner=my_owner,
                timestamp=timestamp,
                created=timestamp + dt.timedelta(seconds=2),
            )
        # when
        result = self.modeladmin._avg_turnaround_time(my_owner)
        # then
        self.assertEqual(result, "2 | 2 | 2")


class TestStructureAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        cls.modeladmin = StructureAdmin(model=Structure, admin_site=AdminSite())
        cls.user = SuperuserFactory()
        cls.alliance = EveAllianceInfoFactory(
            alliance_id=3001, alliance_name="Wayne Enterprises"
        )
        corporation = EveCorporationInfoFactory(
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            alliance=cls.alliance,
        )
        cls.owner = OwnerFactory(corporation=corporation)

    def test_owner(self):
        # given
        obj = StructureFactory(owner=self.owner)
        # when/then
        self.assertEqual(
            self.modeladmin._owner(obj),
            "Wayne Technologies<br>Wayne Enterprises",
        )

    def test_location_structure(self):
        # given
        obj = StructureFactory(
            owner=self.owner, eve_solar_system_name="Amamake", eve_type_name="Astrahus"
        )
        # when/then
        self.assertEqual(self.modeladmin._location(obj), "Amamake<br>Heimatar")

    def test_location_poco(self):
        # given
        obj = PocoFactory(owner=self.owner, eve_planet_name="Amamake V")
        # when/then
        self.assertEqual(self.modeladmin._location(obj), "Amamake V<br>Heimatar")

    def test_location_starbase(self):
        # given
        obj = StarbaseFactory(owner=self.owner, eve_moon_name="Amamake II - Moon 1")
        self.assertEqual(
            self.modeladmin._location(obj), "Amamake II - Moon 1<br>Heimatar"
        )

    def test_type(self):
        # given
        obj = StructureFactory(owner=self.owner, eve_type_name="Astrahus")
        # when/then
        self.assertEqual(self.modeladmin._type(obj), "Astrahus<br>Citadel")

    def test_tags_1(self):
        # given
        obj = StructureFactory(
            owner=self.owner,
            eve_solar_system_name="Amamake",
            tags=[StructureTagFactory(name="my_tag")],
        )
        # when/then
        self.assertSetEqual(set(self.modeladmin._tags(obj)), {"lowsec", "my_tag"})

    def test_tags_2(self):
        # given
        obj = StructureFactory(owner=self.owner)
        obj.tags.clear()
        # when/then
        self.assertListEqual(self.modeladmin._tags(obj), [])

    @patch(MODULE_PATH + ".StructureAdmin.message_user", spec=True)
    def test_action_add_default_tags(self, mock_message_user):
        # given
        obj = StructureFactory()
        obj.tags.clear()
        queryset = Structure.objects.all()
        # when
        self.modeladmin.add_default_tags(MockRequest(self.user), queryset)
        default_tags = StructureTag.objects.filter(is_default=True)
        for obj in queryset:
            self.assertSetEqual(set(obj.tags.all()), set(default_tags))
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".StructureAdmin.message_user", spec=True)
    def test_action_remove_user_tags(self, mock_message_user):
        # given
        obj = StructureFactory(tags=[StructureTagFactory(name="my_tag")])
        queryset = Structure.objects.all()
        # when
        self.modeladmin.remove_user_tags(MockRequest(self.user), queryset)
        for obj in queryset:
            self.assertFalse(obj.tags.filter(is_user_managed=True).exists())
        self.assertTrue(mock_message_user.called)

    def test_owner_corporations_status_filter(self):
        class StructureAdminTest(admin.ModelAdmin):
            list_filter = (OwnerCorporationsFilter,)

        OwnerFactory(
            corporation=EveCorporationInfoFactory(
                corporation_id=2002, corporation_name="Wayne Foods"
            )
        )
        my_modeladmin = StructureAdminTest(Structure, AdminSite())

        # Make sure the lookups are correct
        request = self.factory.get("/")
        request.user = self.user
        changelist = my_modeladmin.get_changelist_instance(request)
        filterspec = changelist.get_filters(request)[0][0]
        expected = [(2002, "Wayne Foods"), (2001, "Wayne Technologies")]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned
        request = self.factory.get("/", {"owner_corporation_id__exact": 2001})
        request.user = self.user
        changelist = my_modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = Structure.objects.filter(owner=self.owner)
        self.assertSetEqual(set(queryset), set(expected))

    def test_owner_alliance_status_filter(self):
        class StructureAdminTest(admin.ModelAdmin):
            list_filter = (OwnerAllianceFilter,)

        # create test data
        owner_2002 = OwnerFactory(
            corporation=EveCorporationInfoFactory(
                corporation_id=2002, alliance=self.alliance
            )
        )
        OwnerFactory(
            corporation=EveCorporationInfoFactory(
                corporation_id=2102, create_alliance=False
            )
        )
        modeladmin = StructureAdminTest(Structure, AdminSite())

        # Make sure the lookups are correct
        request = self.factory.get("/")
        request.user = self.user
        changelist = modeladmin.get_changelist_instance(request)
        filterspec = changelist.get_filters(request)[0][0]
        expected = [(3001, "Wayne Enterprises")]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned
        request = self.factory.get("/", {"owner_alliance_id__exact": 3001})
        request.user = self.user
        changelist = modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = Structure.objects.filter(owner__in=[self.owner, owner_2002])
        self.assertSetEqual(set(queryset), set(expected))


class TestWebhookAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.modeladmin = WebhookAdmin(model=Webhook, admin_site=AdminSite())
        cls.user = SuperuserFactory()

    @patch(MODULE_PATH + ".WebhookAdmin.message_user", spec=True)
    @patch(MODULE_PATH + ".tasks.send_test_notifications_to_webhook")
    def test_action_test_notification(self, mock_task, mock_message_user):
        # given
        WebhookFactory()
        queryset = Webhook.objects.all()
        # when
        self.modeladmin.test_notification(MockRequest(self.user), queryset)
        # then
        self.assertEqual(mock_task.delay.call_count, 1)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".WebhookAdmin.message_user", spec=True)
    def test_action_activate(self, mock_message_user):
        # given
        webhook = WebhookFactory(is_active=False)
        queryset = Webhook.objects.all()
        # when
        self.modeladmin.activate(MockRequest(self.user), queryset)
        # then
        webhook.refresh_from_db()
        self.assertTrue(webhook.is_active)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".WebhookAdmin.message_user", spec=True)
    def test_action_deactivate(self, mock_message_user):
        # given
        webhook = WebhookFactory(is_active=True)
        queryset = Webhook.objects.all()
        # when
        self.modeladmin.deactivate(MockRequest(self.user), queryset)
        # then
        webhook.refresh_from_db()
        self.assertFalse(webhook.is_active)
        self.assertTrue(mock_message_user.called)

    def test_can_assign_owner(self):
        # given
        owner = OwnerFactory()
        webhook = WebhookFactory()
        self.client.force_login(self.user)
        data = {
            "name": webhook.name,
            "notification_types": NotificationType.webhook_defaults(),
            "owners": [owner.pk],
            "url": webhook.url,
            "webhook_type": webhook.webhook_type,
        }
        r = self.client.post(
            f"/admin/structures/webhook/{webhook.pk}/change/", data=data
        )
        self.assertEqual(r.status_code, 302)
        webhook.refresh_from_db()
        self.assertIn(owner, webhook.owners.all())



# flake8: noqa
"""scripts generates large amount of random structures for load testing"""

import inspect
import os
import sys

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
myauth_dir = (
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(current_dir))))
    + "/myauth"
)
sys.path.insert(0, myauth_dir)


import django
from django.apps import apps

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

if not apps.is_installed("structures"):
    raise RuntimeError("The app structures is not installed")

from datetime import timedelta
from random import randrange

from tqdm import tqdm

from django.utils.timezone import now
from esi.clients import esi_client_factory
from eveuniverse.models import EveMoon, EvePlanet, EveSolarSystem, EveType

from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo

from structures.constants import EveTypeId
from structures.models import Owner, Structure, StructureService, StructureTag
from structures.tests.testdata.factories import (
    JumpGateFactory,
    PocoFactory,
    StarbaseFactory,
    StructureFactory,
    WebhookFactory,
)

# TODO: Add data for assets, e.g. fittings

print(
    "generate_structure - "
    "scripts generates large amount of random structures for load testing "
)

STRUCTURES_COUNT = 50
POCOS_COUNT = 25
STARBASES_COUNT = 10

# random pick of most active corporations on zKillboard in Jan 2020
corporation_ids = [
    98388312,
    98558506,
    98370861,
    98410772,
    98148549,
    98431483,
    667531913,
    427125032,
    98514543,
    98013740,
]
structure_type_ids = [35825, 35826, 35827, 35832, 35832, 35834, 35835, 35836, 35841]
solar_system_ids = [30000142, 30001445, 30002355, 30004046, 30003833, 30045338]
planet_ids = [
    40477808,
    40393603,
    40423560,
    40024173,
    40189130,
    40133677,
    40161704,
    40045256,
    40026932,
    40147622,
]
moon_ids = [
    40423562,
    40171691,
    40419806,
    40133313,
    40380849,
    40411632,
    40379111,
    40325304,
    40393615,
    40380579,
]
services = [
    "Clone Bay",
    "Moondrilling",
    "Reprocessing",
    "Market Hub",
    "Manufacturing (Standard)",
    "Blueprint Copying",
    "Material Efficiency Research",
    "Time Efficiency Research",
]
tag_names = [
    "Top Secret",
    "Priority",
    "Trash",
    "Needs caretaker",
    "Taskforce Bravo",
    "Not so friendly",
]


def get_random(lst: list) -> object:
    return lst[randrange(len(lst))]


def get_random_subset(lst: list, max_members: int = None) -> list:
    lst2 = lst.copy()
    subset = []
    if not max_members:
        max_members = len(lst)
    else:
        max_members = min(max_members, len(lst))

    for x in range(randrange(max_members) + 1):
        m = lst2.pop(randrange(len(lst2)))
        subset.append(m)

    return subset


print("Connecting to ESI ...")
client = esi_client_factory()

# generating data
webhook = WebhookFactory(name="Generated webhook")
owners = []
for corporation_id in tqdm(corporation_ids, desc="Creating owners"):
    try:
        corporation = client.Corporation.get_corporations_corporation_id(
            corporation_id=corporation_id
        ).result()
        try:
            EveAllianceInfo.objects.get(alliance_id=corporation["alliance_id"])
        except EveAllianceInfo.DoesNotExist:
            EveAllianceInfo.objects.create_alliance(corporation["alliance_id"])
    except Exception:
        pass

    try:
        corporation = EveCorporationInfo.objects.get(corporation_id=corporation_id)
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(corporation_id)

    owner, created = Owner.objects.get_or_create(
        corporation=corporation, defaults={"is_active": False}
    )
    if created:
        owner.webhooks.add(webhook)

    owner.assets_last_update_at = now()
    owner.structures_last_update_at = now()
    owner.save()
    owners.append(owner)

# fetching data from ESI
structure_eve_types = [
    EveType.objects.get_or_create_esi(id=type_id)[0] for type_id in structure_type_ids
]
EveType.objects.get_or_create_esi(id=EveTypeId.CUSTOMS_OFFICE)
starbase_eve_type_ids = [
    12235,
    12236,
    16213,
    16214,
    16286,
    20059,
    20060,
    20061,
    20062,
    20063,
    20064,
    20065,
    20066,
]
starbase_eve_types = [
    EveType.objects.get_or_create_esi(id=type_id)[0]
    for type_id in starbase_eve_type_ids
]
eve_solar_systems = [
    EveSolarSystem.objects.get_or_create_esi(id=system_id)[0]
    for system_id in solar_system_ids
]
eve_planets = [
    EvePlanet.objects.get_or_create_esi(id=system_id)[0] for system_id in planet_ids
]
eve_moons = [
    EveMoon.objects.get_or_create_esi(id=system_id)[0] for system_id in moon_ids
]

# generate random tags
tags = []
for name in tag_names:
    tag, _ = StructureTag.objects.update_or_create(
        name=name,
        defaults={"style": get_random([x[0] for x in StructureTag.Style.choices])},
    )
    tags.append(tag)

# remove old structures
Structure.objects.filter(owner__in=owners).delete()

# creating upwell structures
for i in tqdm(range(1, STRUCTURES_COUNT + 1), desc="Creating upwell structures"):
    eve_type = get_random(structure_eve_types)
    eve_solar_system = get_random(eve_solar_systems)
    reinforce_hour = randrange(24)
    owner = get_random(owners)
    state = get_random(
        [
            Structure.State.SHIELD_VULNERABLE,
            Structure.State.SHIELD_VULNERABLE,
            Structure.State.SHIELD_VULNERABLE,
            Structure.State.SHIELD_VULNERABLE,
            Structure.State.SHIELD_VULNERABLE,
            Structure.State.SHIELD_VULNERABLE,
            Structure.State.ARMOR_REINFORCE,
            Structure.State.HULL_REINFORCE,
        ]
    )
    is_low_power = get_random([True, False]) or state == Structure.State.HULL_REINFORCE

    unanchors_at = None

    if not is_low_power:
        fuel_expires_at = now() + timedelta(days=randrange(14), hours=randrange(12))
        last_online_at = now()
        if randrange(1, 10) == 1:
            unanchors_at = now() + timedelta(days=randrange(6), hours=randrange(12))
    else:
        fuel_expires_at = None
        last_online_at = now() - timedelta(days=get_random([1, 2, 3, 10]))

    params = {
        "owner": owner,
        "name": f"Generated structure #{i:05d}",
        "eve_solar_system": eve_solar_system,
        "reinforce_hour": reinforce_hour,
        "state": state,
        "fuel_expires_at": fuel_expires_at,
        "last_online_at": last_online_at,
        "unanchors_at": unanchors_at,
    }
    if eve_type.id == EveTypeId.JUMP_GATE:
        structure = JumpGateFactory(**params)

    else:
        structure = StructureFactory(**{**params, **{"eve_type": eve_type}})
        if is_low_power:
            state = StructureService.State.OFFLINE
        else:
            state = StructureService.State.ONLINE
        for name in get_random_subset(services, 3):
            StructureService.objects.create(structure=structure, name=name, state=state)

    if structure.is_reinforced:
        state_timer_start = now() - timedelta(days=randrange(3), hours=randrange(12))
        state_timer_end = now() + timedelta(days=randrange(3), hours=randrange(12))
        structure.state_timer_start = state_timer_start
        structure.state_timer_end = state_timer_end
        structure.save()

    # common
    structure.tags.add(*get_random_subset(tags))


# creating pocos
for i in tqdm(range(1, POCOS_COUNT + 1), desc="Creating pocos"):
    eve_planet = get_random(eve_planets)
    reinforce_hour = randrange(24)
    owner = get_random(owners)

    structure = PocoFactory(
        owner=owner, eve_planet=eve_planet, reinforce_hour=reinforce_hour
    )

    structure.tags.add(*get_random_subset(tags))

# creating starbases
for i in tqdm(range(1, STARBASES_COUNT + 1), desc="Creating starbases"):
    eve_type = get_random(starbase_eve_types)
    eve_moon = get_random(eve_moons)
    reinforce_hour = randrange(24)
    owner = get_random(owners)
    state = get_random(
        [
            Structure.State.POS_ONLINE,
            Structure.State.POS_ONLINE,
            Structure.State.POS_ONLINE,
            Structure.State.POS_ONLINE,
            Structure.State.POS_ONLINE,
            Structure.State.POS_ONLINE,
            Structure.State.POS_ONLINING,
            Structure.State.POS_OFFLINE,
            Structure.State.POS_REINFORCED,
            Structure.State.POS_UNANCHORING,
        ]
    )

    structure = StarbaseFactory(owner=owner, eve_type=eve_type, eve_moon=eve_moon)

    if structure.is_reinforced:
        state_timer_start = now() - timedelta(days=randrange(3), hours=randrange(12))
        state_timer_end = now() + timedelta(days=randrange(3), hours=randrange(12))
        structure.state_timer_start = state_timer_start
        structure.state_timer_end = state_timer_end
        structure.save()

    # common
    structure.tags.add(*get_random_subset(tags))


print("DONE")



"""Upwell structures embeds."""

# pylint: disable=missing-class-docstring

from collections import namedtuple

import dhooks_lite

from django.utils.translation import gettext as _

from app_utils.datetime import ldap_time_2_datetime, ldap_timedelta_2_timedelta

from structures.helpers import get_or_create_eve_entity, get_or_create_eve_type
from structures.models import Notification, Structure, Webhook

from .helpers import (
    gen_corporation_link,
    gen_solar_system_text,
    target_datetime_formatted,
    timeuntil,
)
from .main import NotificationBaseEmbed


class NotificationStructureEmbed(NotificationBaseEmbed):
    """Base class for most structure related notification embeds."""

    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        try:
            structure = Structure.objects.select_related_defaults().get(
                id=self._data["structureID"]
            )
        except Structure.DoesNotExist:
            structure = None
            structure_name = _("(unknown)")
            structure_type = self._notification.eve_structure_type()
            structure_solar_system = self._notification.eve_solar_system(
                "solarsystemID"
            )
            owner_link = "(unknown)"
            location = ""
        else:
            structure_name = structure.name
            structure_type = structure.eve_type
            structure_solar_system = structure.eve_solar_system
            owner_link = gen_corporation_link(str(structure.owner))
            # Translators: Context is a location, e.g. a moon
            location = _(" at %s ") % structure.eve_moon if structure.eve_moon else ""

        self._structure = structure
        self._description = (
            _(
                "The %(structure_type)s %(structure_name)s%(location)s "
                "in %(solar_system)s belonging to %(owner_link)s"
            )
            % {
                "structure_type": structure_type.name,
                "structure_name": Webhook.text_bold(structure_name),
                "location": location,
                "solar_system": gen_solar_system_text(structure_solar_system),
                "owner_link": owner_link,
            }
            + " "
        )
        self._thumbnail = dhooks_lite.Thumbnail(
            structure_type.icon_url(size=self.ICON_DEFAULT_SIZE)
        )


class NotificationStructureOnline(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Structure online")
        self._description += _("is now online.")
        self._color = Webhook.Color.SUCCESS


class NotificationStructureFuelAlert(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        if self._structure and self._structure.fuel_expires_at:
            hours_left = timeuntil(self._structure.fuel_expires_at)
        else:
            hours_left = "?"
        self._title = _("Structure fuel alert")
        self._description += _("is running out of fuel in %s.") % Webhook.text_bold(
            hours_left
        )

        self._color = Webhook.Color.WARNING


class NotificationStructureJumpFuelAlert(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Jump gate low on Liquid Ozone")
        threshold_str = f"{self._data['threshold']:,}"
        quantity_str = f"{self._structure.jump_fuel_quantity():,}"
        self._description += _(
            "is below %(threshold)s units on Liquid Ozone.\n"
            "Remaining units: %(remaining)s."
        ) % {
            "threshold": f"{Webhook.text_bold(threshold_str)}",
            "remaining": f"{Webhook.text_bold(quantity_str)}",
        }
        self._color = Webhook.Color.WARNING


class NotificationStructureRefueledExtra(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        target_date = self.fuel_expires_target_date()
        self._title = _("Structure refueled")
        self._description += (
            _("has been refueled. Fuel will last until %s.") % target_date
        )
        self._color = Webhook.Color.INFO


class NotificationStructureServicesOffline(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Structure services off-line")
        self._description += _("has all services off-lined.")
        if self._structure and self._structure.services.count() > 0:
            qs = self._structure.services.all().order_by("name")
            services_list = "\n".join([x.name for x in qs])
            self._description += f"\n*{services_list}*"
        self._color = Webhook.Color.DANGER


class NotificationStructureWentLowPower(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Structure low power")
        self._description += _("went to low power mode.")
        self._color = Webhook.Color.WARNING


class NotificationStructureWentHighPower(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Structure full power")
        self._description += _("went to full power mode.")
        self._color = Webhook.Color.SUCCESS


class NotificationStructureUnanchoring(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Structure un-anchoring")
        unanchored_at = notification.timestamp + ldap_timedelta_2_timedelta(
            self._data["timeLeft"]
        )
        self._description += _(
            "has started un-anchoring. It will be fully un-anchored at: %s"
        ) % target_datetime_formatted(unanchored_at)
        self._color = Webhook.Color.INFO


class NotificationStructureUnderAttack(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Structure under attack")
        self._description += _("is under attack by %(attacker)s.\n%(damage_text)s") % {
            "attacker": self.gen_attacker_link(),
            "damage_text": self.compile_damage_text("Percentage"),
        }
        self._color = Webhook.Color.DANGER


class NotificationStructureLostShield(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Structure lost shield")
        timer_ends_at = notification.timestamp + ldap_timedelta_2_timedelta(
            self._data["timeLeft"]
        )
        self._description += _(
            "has lost its shields. Armor timer end at: %s"
        ) % target_datetime_formatted(timer_ends_at)
        self._color = Webhook.Color.DANGER


class NotificationStructureLostArmor(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Structure lost armor")
        timer_ends_at = notification.timestamp + ldap_timedelta_2_timedelta(
            self._data["timeLeft"]
        )
        self._description += _(
            "has lost its armor. Hull timer end at: %s"
        ) % target_datetime_formatted(timer_ends_at)
        self._color = Webhook.Color.DANGER


class NotificationStructureDestroyed(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Structure destroyed")
        self._description += _("has been destroyed.")
        self._color = Webhook.Color.DANGER


class NotificationStructureOwnershipTransferred(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        structure_type = self._notification.eve_structure_type()
        self._description = _(
            "The %(structure_type)s %(structure_name)s in %(solar_system)s "
        ) % {
            "structure_type": structure_type.name,
            "structure_name": Webhook.text_bold(self._data["structureName"]),
            "solar_system": gen_solar_system_text(
                self._notification.eve_solar_system()
            ),
        }
        from_corporation = get_or_create_eve_entity(id=self._data["oldOwnerCorpID"])
        to_corporation = get_or_create_eve_entity(id=self._data["newOwnerCorpID"])
        character = get_or_create_eve_entity(id=self._data["charID"])
        self._description += _(
            "has been transferred from %(from_corporation)s "
            "to %(to_corporation)s by %(character)s."
        ) % {
            "from_corporation": gen_corporation_link(from_corporation.name),
            "to_corporation": gen_corporation_link(to_corporation.name),
            "character": character.name,
        }
        self._title = _("Ownership transferred")
        self._color = Webhook.Color.INFO
        self._thumbnail = dhooks_lite.Thumbnail(
            structure_type.icon_url(size=self.ICON_DEFAULT_SIZE)
        )


class NotificationStructureAnchoring(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        structure_type = self._notification.eve_structure_type()
        solar_system = self._notification.eve_solar_system("solarsystemID")
        owner_link = gen_corporation_link(self._data.get("ownerCorpName", "(unknown)"))
        self._description = _(
            "A %(structure_type)s belonging to %(owner_link)s "
            "has started anchoring in %(solar_system)s. "
        ) % {
            "structure_type": structure_type.name,
            "owner_link": owner_link,
            "solar_system": gen_solar_system_text(solar_system),
        }
        self._title = _("Structure anchoring")
        self._color = Webhook.Color.INFO
        self._thumbnail = dhooks_lite.Thumbnail(
            structure_type.icon_url(size=self.ICON_DEFAULT_SIZE)
        )


class NotificationStructureReinforceChange(NotificationBaseEmbed):
    StructureInfo = namedtuple(
        "StructureInfo", ["name", "eve_type", "eve_solar_system", "owner_link"]
    )

    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        all_structure_info = []
        for structure_info in self._data["allStructureInfo"]:
            try:
                structure = Structure.objects.select_related_defaults().get(
                    id=structure_info[0]
                )
            except Structure.DoesNotExist:
                all_structure_info.append(
                    self.StructureInfo(
                        name=structure_info[1],
                        eve_type=get_or_create_eve_type(id=structure_info[2]),
                        eve_solar_system=None,
                        owner_link=_("(unknown)"),
                    )
                )
            else:
                all_structure_info.append(
                    self.StructureInfo(
                        name=structure.name,
                        eve_type=structure.eve_type,
                        eve_solar_system=structure.eve_solar_system,
                        owner_link=gen_corporation_link(str(structure.owner)),
                    )
                )

        self._title = _("Structure reinforcement time changed")
        change_effective = ldap_time_2_datetime(self._data["timestamp"])
        self._description = _(
            "Reinforcement hour has been changed to %s "
            "for the following structures:\n"
        ) % Webhook.text_bold(self._data["hour"])
        for structure_info in all_structure_info:
            if eve_solar_system := structure_info.eve_solar_system:
                solar_system_text = gen_solar_system_text(eve_solar_system)
            else:
                solar_system_text = ""
            self._description += _(
                "- %(structure_type)s %(structure_name)s in %(solar_system)s "
                "belonging to %(owner_link)s"
            ) % {
                "structure_type": structure_info.eve_type.name,
                "structure_name": Webhook.text_bold(structure_info.name),
                "solar_system": solar_system_text,
                "owner_link": structure_info.owner_link,
            }

        self._description += _(
            _("\n\nChange becomes effective at %s.")
        ) % target_datetime_formatted(change_effective)
        self._color = Webhook.Color.INFO


class NotificationStructureLowReagentsAlert(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Structure low on reagents")
        self._description += _(
            "has insufficient reagents present to support "
            "one full day of moon drill activity."
        )
        self._color = Webhook.Color.WARNING


class NotificationStructureNoReagentsAlert(NotificationStructureEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Structure out of reagents")
        self._description += _(
            "has run out of reagents, "
            "which is preventing further moon material production."
        )
        self._color = Webhook.Color.DANGER
